import iris
import iris.quickplot as qplt
import iris.analysis
import matplotlib.pyplot as plt
import numpy as np


fname = iris.sample_data_path("hybrid_height.nc")
column = iris.load_cube(fname, "air_potential_temperature")[:, 0, 0]

alt_coord = column.coord("altitude")

# Interpolate the "perfect" linear interpolation. Really this is just
# a high number of interpolation points, in this case 1000 of them.
altitude_points = [("altitude", np.linspace(400, 1250, 1000))]
scheme = iris.analysis.Linear(extrapolation_mode="mask")
linear_column = column.interpolate(altitude_points, scheme)

# Now interpolate the data onto 10 evenly spaced altitude levels,
# as we did in the example.
altitude_points = [("altitude", np.linspace(400, 1250, 10))]
scheme = iris.analysis.Linear()
new_column = column.interpolate(altitude_points, scheme)

plt.figure(figsize=(5, 4), dpi=100)

# Plot the black markers for the original data.
qplt.plot(
    column,
    column.coord("altitude"),
    marker="o",
    color="black",
    linestyle="",
    markersize=3,
    label="Original values",
    zorder=2,
)

# Plot the gray line to display the linear interpolation.
qplt.plot(
    linear_column,
    linear_column.coord("altitude"),
    color="gray",
    label="Linear interpolation",
    zorder=0,
)

# Plot the red markers for the new data.
qplt.plot(
    new_column,
    new_column.coord("altitude"),
    marker="D",
    color="red",
    linestyle="",
    label="Interpolated values",
    zorder=1,
)

ax = plt.gca()
# Space the plot such that the labels appear correctly.
plt.subplots_adjust(left=0.17, bottom=0.14)

# Limit the plot to a maximum of 5 ticks.
ax.xaxis.get_major_locator().set_params(nbins=5)

# Prevent matplotlib from using "offset" notation on the xaxis.
ax.xaxis.get_major_formatter().set_useOffset(False)

# Put some space between the line and the axes.
ax.margins(0.05)

# Place gridlines and a legend.
ax.grid()
plt.legend(loc="lower right")

plt.show()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Test function :func:`iris._lazy data.as_concrete_data`."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import numpy as np
import numpy.ma as ma

from iris._lazy_data import as_concrete_data, as_lazy_data, is_lazy_data


class MyProxy:
    def __init__(self, a):
        self.shape = a.shape
        self.dtype = a.dtype
        self.ndim = a.ndim
        self.a = a

    def __getitem__(self, keys):
        return self.a[keys]


class Test_as_concrete_data(tests.IrisTest):
    def test_concrete_input_data(self):
        data = np.arange(24).reshape((4, 6))
        result = as_concrete_data(data)
        self.assertIs(data, result)
        self.assertFalse(is_lazy_data(result))

    def test_concrete_masked_input_data(self):
        data = ma.masked_array([10, 12, 8, 2], mask=[True, True, False, True])
        result = as_concrete_data(data)
        self.assertIs(data, result)
        self.assertFalse(is_lazy_data(result))

    def test_lazy_data(self):
        data = np.arange(24).reshape((2, 12))
        lazy_array = as_lazy_data(data)
        self.assertTrue(is_lazy_data(lazy_array))
        result = as_concrete_data(lazy_array)
        self.assertFalse(is_lazy_data(result))
        self.assertArrayEqual(result, data)

    def test_lazy_mask_data(self):
        data = np.arange(24).reshape((2, 12))
        fill_value = 1234
        mask_data = ma.masked_array(data, fill_value=fill_value)
        lazy_array = as_lazy_data(mask_data)
        self.assertTrue(is_lazy_data(lazy_array))
        result = as_concrete_data(lazy_array)
        self.assertFalse(is_lazy_data(result))
        self.assertMaskedArrayEqual(result, mask_data)
        self.assertEqual(result.fill_value, fill_value)

    def test_lazy_scalar_proxy(self):
        a = np.array(5)
        proxy = MyProxy(a)
        lazy_array = as_lazy_data(proxy)
        self.assertTrue(is_lazy_data(lazy_array))
        result = as_concrete_data(lazy_array)
        self.assertFalse(is_lazy_data(result))
        self.assertEqual(result, a)

    def test_lazy_scalar_proxy_masked(self):
        a = np.ma.masked_array(5, True)
        proxy = MyProxy(a)
        lazy_array = as_lazy_data(proxy)
        self.assertTrue(is_lazy_data(lazy_array))
        result = as_concrete_data(lazy_array)
        self.assertFalse(is_lazy_data(result))
        self.assertMaskedArrayEqual(result, a)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.

"""
Provides context managers which are fundamental to the ability
to run the gallery tests.

"""

import contextlib
import os.path
import warnings
import sys

import matplotlib.pyplot as plt

import iris
from iris._deprecation import IrisDeprecation
import iris.plot as iplt
import iris.quickplot as qplt


GALLERY_DIRECTORY = os.path.join(
    os.path.dirname(os.path.dirname(__file__)), "gallery_code"
)
GALLERY_DIRECTORIES = [
    os.path.join(GALLERY_DIRECTORY, the_dir)
    for the_dir in os.listdir(GALLERY_DIRECTORY)
]


@contextlib.contextmanager
def add_gallery_to_path():
    """
    Creates a context manager which can be used to add the iris gallery
    to the PYTHONPATH. The gallery entries are only importable throughout the lifetime
    of this context manager.

    """
    orig_sys_path = sys.path
    sys.path = sys.path[:]
    sys.path += GALLERY_DIRECTORIES
    yield
    sys.path = orig_sys_path


@contextlib.contextmanager
def show_replaced_by_check_graphic(test_case):
    """
    Creates a context manager which can be used to replace the functionality
    of matplotlib.pyplot.show with a function which calls the check_graphic
    method on the given test_case (iris.tests.IrisTest.check_graphic).

    """

    def replacement_show():
        # form a closure on test_case and tolerance
        test_case.check_graphic()

    orig_show = plt.show
    plt.show = iplt.show = qplt.show = replacement_show
    yield
    plt.show = iplt.show = qplt.show = orig_show


@contextlib.contextmanager
def fail_any_deprecation_warnings():
    """
    Create a context in which any deprecation warning will cause an error.

    The context also resets all the iris.FUTURE settings to the defaults, as
    otherwise changes made in one test can affect subsequent ones.

    """
    with warnings.catch_warnings():
        # Detect and error all and any Iris deprecation warnings.
        warnings.simplefilter("error", IrisDeprecation)
        # Run with all default settings in iris.FUTURE.
        default_future_kwargs = iris.Future().__dict__.copy()
        for dead_option in iris.Future.deprecated_options:
            # Avoid a warning when setting these !
            del default_future_kwargs[dead_option]
        with iris.FUTURE.context(**default_future_kwargs):
            yield



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.

# import iris tests first so that some things can be initialised before
# importing anything else
import iris.tests as tests

import numpy as np

import iris.cube


class Test_CubeList_getitem(tests.IrisTest):
    def setUp(self):
        self.cube0 = iris.cube.Cube(0)
        self.cube1 = iris.cube.Cube(1)
        self.src_list = [self.cube0, self.cube1]
        self.cube_list = iris.cube.CubeList(self.src_list)

    def test_single(self):
        # Check that simple indexing returns the relevant member Cube.
        for i, cube in enumerate(self.src_list):
            self.assertIs(self.cube_list[i], cube)

    def _test_slice(self, keys):
        subset = self.cube_list[keys]
        self.assertIsInstance(subset, iris.cube.CubeList)
        self.assertEqual(subset, self.src_list[keys])

    def test_slice(self):
        # Check that slicing returns a CubeList containing the relevant
        # members.
        self._test_slice(slice(None))
        self._test_slice(slice(1))
        self._test_slice(slice(1, None))
        self._test_slice(slice(0, 1))
        self._test_slice(slice(None, None, -1))


class Test_CubeList_getslice(tests.IrisTest):
    def setUp(self):
        self.cube0 = iris.cube.Cube(0)
        self.cube1 = iris.cube.Cube(1)
        self.src_list = [self.cube0, self.cube1]
        self.cube_list = iris.cube.CubeList(self.src_list)

    def _test_slice(self, cube_list, equivalent):
        self.assertIsInstance(cube_list, iris.cube.CubeList)
        self.assertEqual(cube_list, equivalent)

    def test_slice(self):
        # Check that slicing returns a CubeList containing the relevant
        # members.
        # NB. We have to use explicit [:1] syntax to trigger the call
        # to __getslice__. Using [slice(1)] still calls __getitem__!
        self._test_slice(self.cube_list[:1], self.src_list[:1])
        self._test_slice(self.cube_list[1:], self.src_list[1:])
        self._test_slice(self.cube_list[0:1], self.src_list[0:1:])


class Test_Cube_add_dim_coord(tests.IrisTest):
    def setUp(self):
        self.cube = iris.cube.Cube(np.arange(4).reshape(2, 2))

    def test_no_dim(self):
        self.assertRaises(
            TypeError,
            self.cube.add_dim_coord,
            iris.coords.DimCoord(np.arange(2), "latitude"),
        )

    def test_adding_aux_coord(self):
        coord = iris.coords.AuxCoord(np.arange(2), "latitude")
        with self.assertRaises(ValueError):
            self.cube.add_dim_coord(coord, 0)


class TestEquality(tests.IrisTest):
    def test_not_implmemented(self):
        class Terry:
            pass

        cube = iris.cube.Cube(0)
        self.assertIs(cube.__eq__(Terry()), NotImplemented)
        self.assertIs(cube.__ne__(Terry()), NotImplemented)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for :func:`iris.fileformats.rules._make_cube`."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from unittest import mock
import warnings

import numpy as np

from iris.fileformats.rules import _make_cube
from iris.fileformats.rules import ConversionMetadata


class Test(tests.IrisTest):
    def test_invalid_units(self):
        # Mock converter() function that returns an invalid
        # units string amongst the collection of other elements.
        factories = None
        references = None
        standard_name = None
        long_name = None
        units = "wibble"  # Invalid unit.
        attributes = dict(source="test")
        cell_methods = None
        dim_coords_and_dims = None
        aux_coords_and_dims = None
        metadata = ConversionMetadata(
            factories,
            references,
            standard_name,
            long_name,
            units,
            attributes,
            cell_methods,
            dim_coords_and_dims,
            aux_coords_and_dims,
        )
        converter = mock.Mock(return_value=metadata)

        data = np.arange(3.0)
        field = mock.Mock(
            core_data=lambda: data, bmdi=9999.0, realised_dtype=data.dtype
        )
        with warnings.catch_warnings(record=True) as warn:
            warnings.simplefilter("always")
            cube, factories, references = _make_cube(field, converter)

        # Check attributes dictionary is correctly populated.
        expected_attributes = attributes.copy()
        expected_attributes["invalid_units"] = units
        self.assertEqual(cube.attributes, expected_attributes)

        # Check warning was raised.
        self.assertEqual(len(warn), 1)
        exp_emsg = "invalid units {!r}".format(units)
        self.assertRegex(str(warn[0]), exp_emsg)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Provide conversion to and from Pandas data structures.

See also: http://pandas.pydata.org/

"""

import datetime

import cf_units
from cf_units import Unit
import cftime
import numpy as np
import numpy.ma as ma
import pandas

try:
    from pandas.core.indexes.datetimes import DatetimeIndex  # pandas >=0.20
except ImportError:
    from pandas.tseries.index import DatetimeIndex  # pandas <0.20

import iris
from iris.coords import AuxCoord, DimCoord
from iris.cube import Cube


def _add_iris_coord(cube, name, points, dim, calendar=None):
    """
    Add a Coord to a Cube from a Pandas index or columns array.

    If no calendar is specified for a time series, Gregorian is assumed.

    """
    units = Unit("unknown")
    if calendar is None:
        calendar = cf_units.CALENDAR_GREGORIAN

    # Convert pandas datetime objects to python datetime obejcts.
    if isinstance(points, DatetimeIndex):
        points = np.array([i.to_pydatetime() for i in points])

    # Convert datetime objects to Iris' current datetime representation.
    if points.dtype == object:
        dt_types = (datetime.datetime, cftime.datetime)
        if all([isinstance(i, dt_types) for i in points]):
            units = Unit("hours since epoch", calendar=calendar)
            points = units.date2num(points)

    points = np.array(points)
    if np.issubdtype(points.dtype, np.number) and iris.util.monotonic(
        points, strict=True
    ):
        coord = DimCoord(points, units=units)
        coord.rename(name)
        cube.add_dim_coord(coord, dim)
    else:
        coord = AuxCoord(points, units=units)
        coord.rename(name)
        cube.add_aux_coord(coord, dim)


def as_cube(pandas_array, copy=True, calendars=None):
    """
    Convert a Pandas array into an Iris cube.

    Args:

        * pandas_array - A Pandas Series or DataFrame.

    Kwargs:

        * copy      - Whether to make a copy of the data.
                      Defaults to True.

        * calendars - A dict mapping a dimension to a calendar.
                      Required to convert datetime indices/columns.

    Example usage::

        as_cube(series, calendars={0: cf_units.CALENDAR_360_DAY})
        as_cube(data_frame, calendars={1: cf_units.CALENDAR_GREGORIAN})

    .. note:: This function will copy your data by default.

    """
    calendars = calendars or {}
    if pandas_array.ndim not in [1, 2]:
        raise ValueError(
            "Only 1D or 2D Pandas arrays "
            "can currently be conveted to Iris cubes."
        )

    # Make the copy work consistently across NumPy 1.6 and 1.7.
    # (When 1.7 takes a copy it preserves the C/Fortran ordering, but
    # 1.6 doesn't. Since we don't care about preserving the order we can
    # just force it back to C-order.)
    order = "C" if copy else "A"
    data = np.array(pandas_array, copy=copy, order=order)
    cube = Cube(np.ma.masked_invalid(data, copy=False))
    _add_iris_coord(
        cube, "index", pandas_array.index, 0, calendars.get(0, None)
    )
    if pandas_array.ndim == 2:
        _add_iris_coord(
            cube,
            "columns",
            pandas_array.columns.values,
            1,
            calendars.get(1, None),
        )
    return cube


def _as_pandas_coord(coord):
    """Convert an Iris Coord into a Pandas index or columns array."""
    index = coord.points
    if coord.units.is_time_reference():
        index = coord.units.num2date(index)
    return index


def _assert_shared(np_obj, pandas_obj):
    """Ensure the pandas object shares memory."""
    values = pandas_obj.values

    def _get_base(array):
        # Chase the stack of NumPy `base` references back to the original array
        while array.base is not None:
            array = array.base
        return array

    base = _get_base(values)
    np_base = _get_base(np_obj)
    if base is not np_base:
        msg = "Pandas {} does not share memory".format(
            type(pandas_obj).__name__
        )
        raise AssertionError(msg)


def as_series(cube, copy=True):
    """
    Convert a 1D cube to a Pandas Series.

    Args:

        * cube - The cube to convert to a Pandas Series.

    Kwargs:

        * copy - Whether to make a copy of the data.
                 Defaults to True. Must be True for masked data.

    .. note::

        This function will copy your data by default.
        If you have a large array that cannot be copied,
        make sure it is not masked and use copy=False.

    """
    data = cube.data
    if ma.isMaskedArray(data):
        if not copy:
            raise ValueError("Masked arrays must always be copied.")
        data = data.astype("f").filled(np.nan)
    elif copy:
        data = data.copy()

    index = None
    if cube.dim_coords:
        index = _as_pandas_coord(cube.dim_coords[0])

    series = pandas.Series(data, index)
    if not copy:
        _assert_shared(data, series)

    return series


def as_data_frame(cube, copy=True):
    """
    Convert a 2D cube to a Pandas DataFrame.

    Args:

        * cube - The cube to convert to a Pandas DataFrame.

    Kwargs:

        * copy - Whether to make a copy of the data.
                 Defaults to True. Must be True for masked data
                 and some data types (see notes below).

    .. note::

        This function will copy your data by default.
        If you have a large array that cannot be copied,
        make sure it is not masked and use copy=False.

    .. note::

        Pandas will sometimes make a copy of the array,
        for example when creating from an int32 array.
        Iris will detect this and raise an exception if copy=False.

    """
    data = cube.data
    if ma.isMaskedArray(data):
        if not copy:
            raise ValueError("Masked arrays must always be copied.")
        data = data.astype("f").filled(np.nan)
    elif copy:
        data = data.copy()

    index = columns = None
    if cube.coords(dimensions=[0]):
        index = _as_pandas_coord(cube.coord(dimensions=[0]))
    if cube.coords(dimensions=[1]):
        columns = _as_pandas_coord(cube.coord(dimensions=[1]))

    data_frame = pandas.DataFrame(data, index, columns)
    if not copy:
        _assert_shared(data, data_frame)

    return data_frame



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the :class:`iris.coords.Coord` class."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import collections
from unittest import mock
import warnings

import dask.array as da
import numpy as np

import iris
from iris.cube import Cube
from iris.coords import DimCoord, AuxCoord, Coord
from iris.exceptions import UnitConversionError
from iris.tests.unit.coords import CoordTestMixin


Pair = collections.namedtuple("Pair", "points bounds")


class Test_nearest_neighbour_index__ascending(tests.IrisTest):
    def setUp(self):
        points = [0.0, 90.0, 180.0, 270.0]
        self.coord = DimCoord(points, circular=False, units="degrees")

    def _test_nearest_neighbour_index(
        self, target, bounds=None, circular=False
    ):
        _bounds = [[-20, 10], [10, 100], [100, 260], [260, 340]]
        ext_pnts = [-70, -10, 110, 275, 370]
        if bounds is True:
            self.coord.bounds = _bounds
        else:
            self.coord.bounds = bounds
        self.coord.circular = circular
        results = [self.coord.nearest_neighbour_index(ind) for ind in ext_pnts]
        self.assertEqual(results, target)

    def test_nobounds(self):
        target = [0, 0, 1, 3, 3]
        self._test_nearest_neighbour_index(target)

    def test_nobounds_circular(self):
        target = [3, 0, 1, 3, 0]
        self._test_nearest_neighbour_index(target, circular=True)

    def test_bounded(self):
        target = [0, 0, 2, 3, 3]
        self._test_nearest_neighbour_index(target, bounds=True)

    def test_bounded_circular(self):
        target = [3, 0, 2, 3, 0]
        self._test_nearest_neighbour_index(target, bounds=True, circular=True)

    def test_bounded_overlapping(self):
        _bounds = [[-20, 50], [10, 150], [100, 300], [260, 340]]
        target = [0, 0, 1, 2, 3]
        self._test_nearest_neighbour_index(target, bounds=_bounds)

    def test_bounded_disjointed(self):
        _bounds = [[-20, 10], [80, 170], [180, 190], [240, 340]]
        target = [0, 0, 1, 3, 3]
        self._test_nearest_neighbour_index(target, bounds=_bounds)

    def test_scalar(self):
        self.coord = DimCoord([0], circular=False, units="degrees")
        target = [0, 0, 0, 0, 0]
        self._test_nearest_neighbour_index(target)


class Test_nearest_neighbour_index__descending(tests.IrisTest):
    def setUp(self):
        points = [270.0, 180.0, 90.0, 0.0]
        self.coord = DimCoord(points, circular=False, units="degrees")

    def _test_nearest_neighbour_index(
        self, target, bounds=False, circular=False
    ):
        _bounds = [[340, 260], [260, 100], [100, 10], [10, -20]]
        ext_pnts = [-70, -10, 110, 275, 370]
        if bounds:
            self.coord.bounds = _bounds
        self.coord.circular = circular
        results = [self.coord.nearest_neighbour_index(ind) for ind in ext_pnts]
        self.assertEqual(results, target)

    def test_nobounds(self):
        target = [3, 3, 2, 0, 0]
        self._test_nearest_neighbour_index(target)

    def test_nobounds_circular(self):
        target = [0, 3, 2, 0, 3]
        self._test_nearest_neighbour_index(target, circular=True)

    def test_bounded(self):
        target = [3, 3, 1, 0, 0]
        self._test_nearest_neighbour_index(target, bounds=True)

    def test_bounded_circular(self):
        target = [0, 3, 1, 0, 3]
        self._test_nearest_neighbour_index(target, bounds=True, circular=True)


class Test_guess_bounds(tests.IrisTest):
    def setUp(self):
        self.coord = DimCoord(
            np.array([-160, -120, 0, 30, 150, 170]),
            units="degree",
            standard_name="longitude",
            circular=True,
        )

    def test_non_circular(self):
        self.coord.circular = False
        self.coord.guess_bounds()
        target = np.array(
            [
                [-180.0, -140.0],
                [-140.0, -60.0],
                [-60.0, 15.0],
                [15.0, 90.0],
                [90.0, 160.0],
                [160.0, 180.0],
            ]
        )
        self.assertArrayEqual(target, self.coord.bounds)

    def test_circular_increasing(self):
        self.coord.guess_bounds()
        target = np.array(
            [
                [-175.0, -140.0],
                [-140.0, -60.0],
                [-60.0, 15.0],
                [15.0, 90.0],
                [90.0, 160.0],
                [160.0, 185.0],
            ]
        )
        self.assertArrayEqual(target, self.coord.bounds)

    def test_circular_decreasing(self):
        self.coord.points = self.coord.points[::-1]
        self.coord.guess_bounds()
        target = np.array(
            [
                [185.0, 160.0],
                [160.0, 90.0],
                [90.0, 15.0],
                [15.0, -60.0],
                [-60.0, -140.0],
                [-140.0, -175.0],
            ]
        )
        self.assertArrayEqual(target, self.coord.bounds)

    def test_circular_increasing_alt_range(self):
        self.coord.points = np.array([10, 30, 90, 150, 210, 220])
        self.coord.guess_bounds()
        target = np.array(
            [
                [-65.0, 20.0],
                [20.0, 60.0],
                [60.0, 120.0],
                [120.0, 180.0],
                [180.0, 215.0],
                [215.0, 295.0],
            ]
        )
        self.assertArrayEqual(target, self.coord.bounds)

    def test_circular_decreasing_alt_range(self):
        self.coord.points = np.array([10, 30, 90, 150, 210, 220])[::-1]
        self.coord.guess_bounds()
        target = np.array(
            [
                [295, 215],
                [215, 180],
                [180, 120],
                [120, 60],
                [60, 20],
                [20, -65],
            ]
        )
        self.assertArrayEqual(target, self.coord.bounds)


class Test_guess_bounds__default_enabled_latitude_clipping(tests.IrisTest):
    def test_all_inside(self):
        lat = DimCoord([-10, 0, 20], units="degree", standard_name="latitude")
        lat.guess_bounds()
        self.assertArrayEqual(lat.bounds, [[-15, -5], [-5, 10], [10, 30]])

    def test_points_inside_bounds_outside(self):
        lat = DimCoord([-80, 0, 70], units="degree", standard_name="latitude")
        lat.guess_bounds()
        self.assertArrayEqual(lat.bounds, [[-90, -40], [-40, 35], [35, 90]])

    def test_points_inside_bounds_outside_grid_latitude(self):
        lat = DimCoord(
            [-80, 0, 70], units="degree", standard_name="grid_latitude"
        )
        lat.guess_bounds()
        self.assertArrayEqual(lat.bounds, [[-90, -40], [-40, 35], [35, 90]])

    def test_points_to_edges_bounds_outside(self):
        lat = DimCoord([-90, 0, 90], units="degree", standard_name="latitude")
        lat.guess_bounds()
        self.assertArrayEqual(lat.bounds, [[-90, -45], [-45, 45], [45, 90]])

    def test_points_outside(self):
        lat = DimCoord(
            [-100, 0, 120], units="degree", standard_name="latitude"
        )
        lat.guess_bounds()
        self.assertArrayEqual(lat.bounds, [[-150, -50], [-50, 60], [60, 180]])

    def test_points_inside_bounds_outside_wrong_unit(self):
        lat = DimCoord([-80, 0, 70], units="feet", standard_name="latitude")
        lat.guess_bounds()
        self.assertArrayEqual(lat.bounds, [[-120, -40], [-40, 35], [35, 105]])

    def test_points_inside_bounds_outside_wrong_name(self):
        lat = DimCoord([-80, 0, 70], units="degree", standard_name="longitude")
        lat.guess_bounds()
        self.assertArrayEqual(lat.bounds, [[-120, -40], [-40, 35], [35, 105]])

    def test_points_inside_bounds_outside_wrong_name_2(self):
        lat = DimCoord(
            [-80, 0, 70], units="degree", long_name="other_latitude"
        )
        lat.guess_bounds()
        self.assertArrayEqual(lat.bounds, [[-120, -40], [-40, 35], [35, 105]])


class Test_cell(tests.IrisTest):
    def _mock_coord(self):
        coord = mock.Mock(
            spec=Coord,
            ndim=1,
            points=np.array([mock.sentinel.time]),
            bounds=np.array([[mock.sentinel.lower, mock.sentinel.upper]]),
        )
        return coord

    def test_time_as_object(self):
        # Ensure Coord.cell() converts the point/bound values to
        # "datetime" objects.
        coord = self._mock_coord()
        coord.units.num2date = mock.Mock(
            side_effect=[
                mock.sentinel.datetime,
                (mock.sentinel.datetime_lower, mock.sentinel.datetime_upper),
            ]
        )
        cell = Coord.cell(coord, 0)
        self.assertIs(cell.point, mock.sentinel.datetime)
        self.assertEqual(
            cell.bound,
            (mock.sentinel.datetime_lower, mock.sentinel.datetime_upper),
        )
        self.assertEqual(
            coord.units.num2date.call_args_list,
            [
                mock.call((mock.sentinel.time,)),
                mock.call((mock.sentinel.lower, mock.sentinel.upper)),
            ],
        )


class Test_collapsed(tests.IrisTest, CoordTestMixin):
    def test_serialize(self):
        # Collapse a string AuxCoord, causing it to be serialised.
        string = Pair(
            np.array(["two", "four", "six", "eight"]),
            np.array(
                [
                    ["one", "three"],
                    ["three", "five"],
                    ["five", "seven"],
                    ["seven", "nine"],
                ]
            ),
        )
        string_nobounds = Pair(np.array(["ecks", "why", "zed"]), None)
        string_multi = Pair(
            np.array(["three", "six", "nine"]),
            np.array(
                [
                    ["one", "two", "four", "five"],
                    ["four", "five", "seven", "eight"],
                    ["seven", "eight", "ten", "eleven"],
                ]
            ),
        )

        def _serialize(data):
            return "|".join(str(item) for item in data.flatten())

        for units in ["unknown", "no_unit"]:
            for points, bounds in [string, string_nobounds, string_multi]:
                coord = AuxCoord(points=points, bounds=bounds, units=units)
                collapsed_coord = coord.collapsed()
                self.assertArrayEqual(
                    collapsed_coord.points, _serialize(points)
                )
                if bounds is not None:
                    for index in np.ndindex(bounds.shape[1:]):
                        index_slice = (slice(None),) + tuple(index)
                        self.assertArrayEqual(
                            collapsed_coord.bounds[index_slice],
                            _serialize(bounds[index_slice]),
                        )

    def test_dim_1d(self):
        # Numeric coords should not be serialised.
        coord = DimCoord(
            points=np.array([2, 4, 6, 8]),
            bounds=np.array([[1, 3], [3, 5], [5, 7], [7, 9]]),
        )
        for units in ["unknown", "no_unit", 1, "K"]:
            coord.units = units
            collapsed_coord = coord.collapsed()
            self.assertArrayEqual(
                collapsed_coord.points, np.mean(coord.points)
            )
            self.assertArrayEqual(
                collapsed_coord.bounds,
                [[coord.bounds.min(), coord.bounds.max()]],
            )

    def test_lazy_points(self):
        # Lazy points should stay lazy after collapse.
        coord = AuxCoord(points=da.from_array(np.arange(5), chunks=5))
        collapsed_coord = coord.collapsed()
        self.assertTrue(collapsed_coord.has_lazy_bounds())
        self.assertTrue(collapsed_coord.has_lazy_points())

    def test_numeric_nd(self):
        coord = AuxCoord(
            points=np.array([[1, 2, 4, 5], [4, 5, 7, 8], [7, 8, 10, 11]])
        )

        collapsed_coord = coord.collapsed()
        self.assertArrayEqual(collapsed_coord.points, np.array([6]))
        self.assertArrayEqual(collapsed_coord.bounds, np.array([[1, 11]]))

        # Test partially collapsing one dimension...
        collapsed_coord = coord.collapsed(1)
        self.assertArrayEqual(
            collapsed_coord.points, np.array([3.0, 6.0, 9.0])
        )
        self.assertArrayEqual(
            collapsed_coord.bounds, np.array([[1, 5], [4, 8], [7, 11]])
        )

        # ... and the other
        collapsed_coord = coord.collapsed(0)
        self.assertArrayEqual(collapsed_coord.points, np.array([4, 5, 7, 8]))
        self.assertArrayEqual(
            collapsed_coord.bounds,
            np.array([[1, 7], [2, 8], [4, 10], [5, 11]]),
        )

    def test_numeric_nd_bounds_all(self):
        self.setupTestArrays((3, 4))
        coord = AuxCoord(self.pts_real, bounds=self.bds_real)

        collapsed_coord = coord.collapsed()
        self.assertArrayEqual(collapsed_coord.points, np.array([55]))
        self.assertArrayEqual(collapsed_coord.bounds, np.array([[-2, 112]]))

    def test_numeric_nd_bounds_second(self):
        self.setupTestArrays((3, 4))
        coord = AuxCoord(self.pts_real, bounds=self.bds_real)
        collapsed_coord = coord.collapsed(1)
        self.assertArrayEqual(collapsed_coord.points, np.array([15, 55, 95]))
        self.assertArrayEqual(
            collapsed_coord.bounds, np.array([[-2, 32], [38, 72], [78, 112]])
        )

    def test_numeric_nd_bounds_first(self):
        self.setupTestArrays((3, 4))
        coord = AuxCoord(self.pts_real, bounds=self.bds_real)
        # ... and the other..
        collapsed_coord = coord.collapsed(0)
        self.assertArrayEqual(
            collapsed_coord.points, np.array([40, 50, 60, 70])
        )
        self.assertArrayEqual(
            collapsed_coord.bounds,
            np.array([[-2, 82], [8, 92], [18, 102], [28, 112]]),
        )

    def test_numeric_nd_bounds_last(self):
        self.setupTestArrays((3, 4))
        coord = AuxCoord(self.pts_real, bounds=self.bds_real)
        # ... and again with -ve dimension specification.
        collapsed_coord = coord.collapsed(-1)
        self.assertArrayEqual(collapsed_coord.points, np.array([15, 55, 95]))
        self.assertArrayEqual(
            collapsed_coord.bounds, np.array([[-2, 32], [38, 72], [78, 112]])
        )

    def test_lazy_nd_bounds_all(self):
        self.setupTestArrays((3, 4))
        coord = AuxCoord(self.pts_real, bounds=self.bds_lazy)

        collapsed_coord = coord.collapsed()

        # Note that the new points get recalculated from the lazy bounds
        #  and so end up as lazy
        self.assertTrue(collapsed_coord.has_lazy_points())
        self.assertTrue(collapsed_coord.has_lazy_bounds())

        self.assertArrayEqual(collapsed_coord.points, np.array([55]))
        self.assertArrayEqual(collapsed_coord.bounds, da.array([[-2, 112]]))

    def test_lazy_nd_bounds_second(self):
        self.setupTestArrays((3, 4))
        coord = AuxCoord(self.pts_real, bounds=self.bds_lazy)

        collapsed_coord = coord.collapsed(1)
        self.assertArrayEqual(collapsed_coord.points, np.array([15, 55, 95]))
        self.assertArrayEqual(
            collapsed_coord.bounds, np.array([[-2, 32], [38, 72], [78, 112]])
        )

    def test_lazy_nd_bounds_first(self):
        self.setupTestArrays((3, 4))
        coord = AuxCoord(self.pts_real, bounds=self.bds_lazy)

        collapsed_coord = coord.collapsed(0)
        self.assertArrayEqual(
            collapsed_coord.points, np.array([40, 50, 60, 70])
        )
        self.assertArrayEqual(
            collapsed_coord.bounds,
            np.array([[-2, 82], [8, 92], [18, 102], [28, 112]]),
        )

    def test_lazy_nd_bounds_last(self):
        self.setupTestArrays((3, 4))
        coord = AuxCoord(self.pts_real, bounds=self.bds_lazy)

        collapsed_coord = coord.collapsed(-1)
        self.assertArrayEqual(collapsed_coord.points, np.array([15, 55, 95]))
        self.assertArrayEqual(
            collapsed_coord.bounds, np.array([[-2, 32], [38, 72], [78, 112]])
        )

    def test_lazy_nd_points_and_bounds(self):

        self.setupTestArrays((3, 4))
        coord = AuxCoord(self.pts_lazy, bounds=self.bds_lazy)

        collapsed_coord = coord.collapsed()

        self.assertTrue(collapsed_coord.has_lazy_points())
        self.assertTrue(collapsed_coord.has_lazy_bounds())

        self.assertArrayEqual(collapsed_coord.points, da.array([55]))
        self.assertArrayEqual(collapsed_coord.bounds, da.array([[-2, 112]]))


class Test_is_compatible(tests.IrisTest):
    def setUp(self):
        self.test_coord = AuxCoord([1.0])
        self.other_coord = self.test_coord.copy()

    def test_noncommon_array_attrs_compatible(self):
        # Non-common array attributes should be ok.
        self.test_coord.attributes["array_test"] = np.array([1.0, 2, 3])
        self.assertTrue(self.test_coord.is_compatible(self.other_coord))

    def test_matching_array_attrs_compatible(self):
        # Matching array attributes should be ok.
        self.test_coord.attributes["array_test"] = np.array([1.0, 2, 3])
        self.other_coord.attributes["array_test"] = np.array([1.0, 2, 3])
        self.assertTrue(self.test_coord.is_compatible(self.other_coord))

    def test_different_array_attrs_incompatible(self):
        # Differing array attributes should make coords incompatible.
        self.test_coord.attributes["array_test"] = np.array([1.0, 2, 3])
        self.other_coord.attributes["array_test"] = np.array([1.0, 2, 777.7])
        self.assertFalse(self.test_coord.is_compatible(self.other_coord))


class Test_contiguous_bounds(tests.IrisTest):
    def test_1d_coord_no_bounds_warning(self):
        coord = DimCoord([0, 1, 2], standard_name="latitude")
        msg = (
            "Coordinate 'latitude' is not bounded, guessing contiguous "
            "bounds."
        )
        with warnings.catch_warnings():
            # Cause all warnings to raise Exceptions
            warnings.simplefilter("error")
            with self.assertRaisesRegex(Warning, msg):
                coord.contiguous_bounds()

    def test_2d_coord_no_bounds_error(self):
        coord = AuxCoord(np.array([[0, 0], [5, 5]]), standard_name="latitude")
        emsg = "Guessing bounds of 2D coords is not currently supported"
        with self.assertRaisesRegex(ValueError, emsg):
            coord.contiguous_bounds()

    def test__sanity_check_bounds_call(self):
        coord = DimCoord([5, 15, 25], bounds=[[0, 10], [10, 20], [20, 30]])
        with mock.patch(
            "iris.coords.Coord._sanity_check_bounds"
        ) as bounds_check:
            coord.contiguous_bounds()
        bounds_check.assert_called_once()

    def test_1d_coord(self):
        coord = DimCoord(
            [2, 4, 6],
            standard_name="latitude",
            bounds=[[1, 3], [3, 5], [5, 7]],
        )
        expected = np.array([1, 3, 5, 7])
        result = coord.contiguous_bounds()
        self.assertArrayEqual(result, expected)

    def test_1d_coord_discontiguous(self):
        coord = DimCoord(
            [2, 4, 6],
            standard_name="latitude",
            bounds=[[1, 3], [4, 5], [5, 7]],
        )
        expected = np.array([1, 4, 5, 7])
        result = coord.contiguous_bounds()
        self.assertArrayEqual(result, expected)

    def test_2d_lon_bounds(self):
        coord = AuxCoord(
            np.array([[1, 3], [1, 3]]),
            bounds=np.array(
                [[[0, 2, 2, 0], [2, 4, 4, 2]], [[0, 2, 2, 0], [2, 4, 4, 2]]]
            ),
        )
        expected = np.array([[0, 2, 4], [0, 2, 4], [0, 2, 4]])
        result = coord.contiguous_bounds()
        self.assertArrayEqual(result, expected)

    def test_2d_lat_bounds(self):
        coord = AuxCoord(
            np.array([[1, 1], [3, 3]]),
            bounds=np.array(
                [[[0, 0, 2, 2], [0, 0, 2, 2]], [[2, 2, 4, 4], [2, 2, 4, 4]]]
            ),
        )
        expected = np.array([[0, 0, 0], [2, 2, 2], [4, 4, 4]])
        result = coord.contiguous_bounds()
        self.assertArrayEqual(result, expected)


class Test_is_contiguous(tests.IrisTest):
    def test_no_bounds(self):
        coord = DimCoord([1, 3])
        result = coord.is_contiguous()
        self.assertFalse(result)

    def test__discontiguity_in_bounds_call(self):
        # Check that :meth:`iris.coords.Coord._discontiguity_in_bounds` is
        # called.
        coord = DimCoord([1, 3], bounds=[[0, 2], [2, 4]])
        with mock.patch(
            "iris.coords.Coord._discontiguity_in_bounds"
        ) as discontiguity_check:
            # Discontiguity returns two objects that are unpacked in
            # `coord.is_contiguous`.
            discontiguity_check.return_value = [None, None]
            coord.is_contiguous(rtol=1e-1, atol=1e-3)
        discontiguity_check.assert_called_with(rtol=1e-1, atol=1e-3)


class Test__discontiguity_in_bounds(tests.IrisTest):
    def setUp(self):
        self.points_3by3 = np.array([[1, 2, 3], [1, 2, 3], [1, 2, 3]])
        self.lon_bounds_3by3 = np.array(
            [
                [[0, 2, 2, 0], [2, 4, 4, 2], [4, 6, 6, 4]],
                [[0, 2, 2, 0], [2, 4, 4, 2], [4, 6, 6, 4]],
                [[0, 2, 2, 0], [2, 4, 4, 2], [4, 6, 6, 4]],
            ]
        )
        self.lat_bounds_3by3 = np.array(
            [
                [[0, 0, 2, 2], [0, 0, 2, 2], [0, 0, 2, 2]],
                [[2, 2, 4, 4], [2, 2, 4, 4], [2, 2, 4, 4]],
                [[4, 4, 6, 6], [4, 4, 6, 6], [4, 4, 6, 6]],
            ]
        )

    def test_1d_contiguous(self):
        coord = DimCoord(
            [-20, 0, 20], bounds=[[-30, -10], [-10, 10], [10, 30]]
        )
        contiguous, diffs = coord._discontiguity_in_bounds()
        self.assertTrue(contiguous)
        self.assertArrayEqual(diffs, np.zeros(2))

    def test_1d_discontiguous(self):
        coord = DimCoord([10, 20, 40], bounds=[[5, 15], [15, 25], [35, 45]])
        contiguous, diffs = coord._discontiguity_in_bounds()
        self.assertFalse(contiguous)
        self.assertArrayEqual(diffs, np.array([0, 10]))

    def test_1d_one_cell(self):
        # Test a 1D coord with a single cell.
        coord = DimCoord(20, bounds=[[10, 30]])
        contiguous, diffs = coord._discontiguity_in_bounds()
        self.assertTrue(contiguous)
        self.assertArrayEqual(diffs, np.array([]))

    def test_2d_contiguous_both_dirs(self):
        coord = AuxCoord(self.points_3by3, bounds=self.lon_bounds_3by3)
        contiguous, diffs = coord._discontiguity_in_bounds()
        diffs_along_x, diffs_along_y = diffs
        self.assertTrue(contiguous)
        self.assertTrue(not diffs_along_x.any())
        self.assertTrue(not diffs_along_y.any())

    def test_2d_discontiguous_along_x(self):
        coord = AuxCoord(
            self.points_3by3[:, ::2], bounds=self.lon_bounds_3by3[:, ::2, :]
        )
        contiguous, diffs = coord._discontiguity_in_bounds()
        diffs_along_x, diffs_along_y = diffs
        self.assertFalse(contiguous)
        self.assertArrayEqual(
            diffs_along_x, np.array([True, True, True]).reshape(3, 1)
        )
        self.assertTrue(not diffs_along_y.any())

    def test_2d_discontiguous_along_y(self):
        coord = AuxCoord(
            self.points_3by3[::2, :], bounds=self.lat_bounds_3by3[::2, :, :]
        )
        contiguous, diffs = coord._discontiguity_in_bounds()
        diffs_along_x, diffs_along_y = diffs
        self.assertFalse(contiguous)
        self.assertTrue(not diffs_along_x.any())
        self.assertArrayEqual(diffs_along_y, np.array([[True, True, True]]))

    def test_2d_discontiguous_along_x_and_y(self):
        coord = AuxCoord(
            np.array([[1, 5], [3, 5]]),
            bounds=np.array(
                [[[0, 2, 2, 0], [4, 6, 6, 4]], [[2, 4, 4, 2], [4, 6, 6, 4]]]
            ),
        )
        contiguous, diffs = coord._discontiguity_in_bounds()
        diffs_along_x, diffs_along_y = diffs
        exp_x_diffs = np.array([True, False]).reshape(2, 1)
        exp_y_diffs = np.array([True, False]).reshape(1, 2)
        self.assertFalse(contiguous)
        self.assertArrayEqual(diffs_along_x, exp_x_diffs)
        self.assertArrayEqual(diffs_along_y, exp_y_diffs)

    def test_2d_contiguous_along_x_atol(self):
        coord = AuxCoord(
            self.points_3by3[:, ::2], bounds=self.lon_bounds_3by3[:, ::2, :]
        )
        # Set a high atol that allows small discontiguities.
        contiguous, diffs = coord._discontiguity_in_bounds(atol=5)
        diffs_along_x, diffs_along_y = diffs
        self.assertTrue(contiguous)
        self.assertArrayEqual(
            diffs_along_x, np.array([False, False, False]).reshape(3, 1)
        )
        self.assertTrue(not diffs_along_y.any())

    def test_2d_one_cell(self):
        # Test a 2D coord with a single cell, where the coord has shape (1, 1).
        coord = AuxCoord(
            self.points_3by3[:1, :1], bounds=self.lon_bounds_3by3[:1, :1, :]
        )
        contiguous, diffs = coord._discontiguity_in_bounds()
        diffs_along_x, diffs_along_y = diffs
        expected_diffs = np.array([], dtype=np.int64)
        self.assertTrue(contiguous)
        self.assertArrayEqual(diffs_along_x, expected_diffs.reshape(1, 0))
        self.assertArrayEqual(diffs_along_y, expected_diffs.reshape(0, 1))

    def test_2d_one_cell_along_x(self):
        # Test a 2D coord with a single cell along the x axis, where the coord
        # has shape (2, 1).
        coord = AuxCoord(
            self.points_3by3[:, :1], bounds=self.lat_bounds_3by3[:, :1, :]
        )
        contiguous, diffs = coord._discontiguity_in_bounds()
        diffs_along_x, diffs_along_y = diffs
        self.assertTrue(contiguous)
        self.assertTrue(not diffs_along_x.any())
        self.assertArrayEqual(diffs_along_y, np.array([0, 0]).reshape(2, 1))

    def test_2d_one_cell_along_y(self):
        # Test a 2D coord with a single cell along the y axis, where the coord
        # has shape (1, 2).
        coord = AuxCoord(
            self.points_3by3[:1, :], bounds=self.lon_bounds_3by3[:1, :, :]
        )
        contiguous, diffs = coord._discontiguity_in_bounds()
        diffs_along_x, diffs_along_y = diffs
        self.assertTrue(contiguous)
        self.assertTrue(not diffs_along_x.any())
        self.assertTrue(not diffs_along_y.any())

    def test_2d_contiguous_mod_360(self):
        # Test that longitude coordinates are adjusted by the 360 modulus when
        # calculating the discontiguities in contiguous bounds.
        coord = AuxCoord(
            [[175, -175], [175, -175]],
            standard_name="longitude",
            bounds=np.array(
                [
                    [[170, 180, 180, 170], [-180, -170, -170, -180]],
                    [[170, 180, 180, 170], [-180, -170, -170, -180]],
                ]
            ),
        )
        contiguous, diffs = coord._discontiguity_in_bounds()
        diffs_along_x, diffs_along_y = diffs
        self.assertTrue(contiguous)
        self.assertTrue(not diffs_along_x.any())
        self.assertTrue(not diffs_along_y.any())

    def test_2d_discontiguous_mod_360(self):
        # Test that longitude coordinates are adjusted by the 360 modulus when
        # calculating the discontiguities in contiguous bounds.
        coord = AuxCoord(
            [[175, -175], [175, -175]],
            standard_name="longitude",
            bounds=np.array(
                [
                    [[170, 180, 180, 170], [10, 20, 20, 10]],
                    [[170, 180, 180, 170], [10, 20, 20, 10]],
                ]
            ),
        )
        contiguous, diffs = coord._discontiguity_in_bounds()
        diffs_along_x, diffs_along_y = diffs
        self.assertFalse(contiguous)
        self.assertArrayEqual(diffs_along_x, np.array([[True], [True]]))
        self.assertTrue(not diffs_along_y.any())

    def test_2d_contiguous_mod_360_not_longitude(self):
        # Test that non-longitude coordinates are not adjusted by the 360
        # modulus when calculating the discontiguities in contiguous bounds.
        coord = AuxCoord(
            [[-150, 350], [-150, 350]],
            standard_name="height",
            bounds=np.array(
                [
                    [[-400, 100, 100, -400], [100, 600, 600, 100]],
                    [[-400, 100, 100, -400], [100, 600, 600, 100]],
                ]
            ),
        )
        contiguous, diffs = coord._discontiguity_in_bounds()
        diffs_along_x, diffs_along_y = diffs
        self.assertTrue(contiguous)
        self.assertTrue(not diffs_along_x.any())
        self.assertTrue(not diffs_along_y.any())

    def test_2d_discontiguous_mod_360_not_longitude(self):
        # Test that non-longitude coordinates are not adjusted by the 360
        # modulus when calculating the discontiguities in discontiguous bounds.
        coord = AuxCoord(
            [[-150, 350], [-150, 350]],
            standard_name="height",
            bounds=np.array(
                [
                    [[-400, 100, 100, -400], [200, 600, 600, 200]],
                    [[-400, 100, 100, -400], [200, 600, 600, 200]],
                ]
            ),
        )
        contiguous, diffs = coord._discontiguity_in_bounds()
        diffs_along_x, diffs_along_y = diffs
        self.assertFalse(contiguous)
        self.assertArrayEqual(diffs_along_x, np.array([[True], [True]]))
        self.assertTrue(not diffs_along_y.any())


class Test__sanity_check_bounds(tests.IrisTest):
    def test_coord_1d_2_bounds(self):
        # Check that a 1d coord with 2 bounds does not raise an error.
        coord = iris.coords.DimCoord(
            [0, 1], standard_name="latitude", bounds=[[0, 1], [1, 2]]
        )
        coord._sanity_check_bounds()

    def test_coord_1d_no_bounds(self):
        coord = iris.coords.DimCoord([0, 1], standard_name="latitude")
        emsg = (
            "Contiguous bounds are only defined for 1D coordinates with "
            "2 bounds."
        )
        with self.assertRaisesRegex(ValueError, emsg):
            coord._sanity_check_bounds()

    def test_coord_1d_1_bounds(self):
        coord = iris.coords.DimCoord(
            [0, 1], standard_name="latitude", bounds=np.array([[0], [1]])
        )
        emsg = (
            "Contiguous bounds are only defined for 1D coordinates with "
            "2 bounds."
        )
        with self.assertRaisesRegex(ValueError, emsg):
            coord._sanity_check_bounds()

    def test_coord_2d_4_bounds(self):
        coord = iris.coords.AuxCoord(
            [[0, 0], [1, 1]],
            standard_name="latitude",
            bounds=np.array(
                [[[0, 0, 1, 1], [0, 0, 1, 1]], [[1, 1, 2, 2], [1, 1, 2, 2]]]
            ),
        )
        coord._sanity_check_bounds()

    def test_coord_2d_no_bounds(self):
        coord = iris.coords.AuxCoord(
            [[0, 0], [1, 1]], standard_name="latitude"
        )
        emsg = (
            "Contiguous bounds are only defined for 2D coordinates with "
            "4 bounds."
        )
        with self.assertRaisesRegex(ValueError, emsg):
            coord._sanity_check_bounds()

    def test_coord_2d_2_bounds(self):
        coord = iris.coords.AuxCoord(
            [[0, 0], [1, 1]],
            standard_name="latitude",
            bounds=np.array([[[0, 1], [0, 1]], [[1, 2], [1, 2]]]),
        )
        emsg = (
            "Contiguous bounds are only defined for 2D coordinates with "
            "4 bounds."
        )
        with self.assertRaisesRegex(ValueError, emsg):
            coord._sanity_check_bounds()

    def test_coord_3d(self):
        coord = iris.coords.AuxCoord(
            np.zeros((2, 2, 2)), standard_name="height"
        )
        emsg = (
            "Contiguous bounds are not defined for coordinates with more "
            "than 2 dimensions."
        )
        with self.assertRaisesRegex(ValueError, emsg):
            coord._sanity_check_bounds()


class Test_convert_units(tests.IrisTest):
    def test_convert_unknown_units(self):
        coord = iris.coords.AuxCoord(1, units="unknown")
        emsg = (
            "Cannot convert from unknown units. "
            'The "units" attribute may be set directly.'
        )
        with self.assertRaisesRegex(UnitConversionError, emsg):
            coord.convert_units("degrees")


class Test___str__(tests.IrisTest):
    def test_short_time_interval(self):
        coord = DimCoord(
            [5], standard_name="time", units="days since 1970-01-01"
        )
        expected = (
            "DimCoord([1970-01-06 00:00:00], standard_name='time', "
            "calendar='gregorian')"
        )
        result = coord.__str__()
        self.assertEqual(expected, result)

    def test_short_time_interval__bounded(self):
        coord = DimCoord(
            [5, 6], standard_name="time", units="days since 1970-01-01"
        )
        coord.guess_bounds()
        expected = (
            "DimCoord([1970-01-06 00:00:00, 1970-01-07 00:00:00], "
            "bounds=[[1970-01-05 12:00:00, 1970-01-06 12:00:00],\n"
            "       [1970-01-06 12:00:00, 1970-01-07 12:00:00]], "
            "standard_name='time', calendar='gregorian')"
        )
        result = coord.__str__()
        self.assertEqual(expected, result)

    def test_long_time_interval(self):
        coord = DimCoord(
            [5], standard_name="time", units="years since 1970-01-01"
        )
        expected = "DimCoord([5], standard_name='time', calendar='gregorian')"
        result = coord.__str__()
        self.assertEqual(expected, result)

    def test_long_time_interval__bounded(self):
        coord = DimCoord(
            [5, 6], standard_name="time", units="years since 1970-01-01"
        )
        coord.guess_bounds()
        expected = (
            "DimCoord([5 6], bounds=[[4.5 5.5]\n [5.5 6.5]], "
            "standard_name='time', calendar='gregorian')"
        )
        result = coord.__str__()
        self.assertEqual(expected, result)

    def test_non_time_unit(self):
        coord = DimCoord([1.0])
        expected = repr(coord)
        result = coord.__str__()
        self.assertEqual(expected, result)


class TestClimatology(tests.IrisTest):
    # Variety of tests for the climatological property of a coord.
    # Only using AuxCoord since there is no different behaviour between Aux
    # and DimCoords for this property.

    def test_create(self):
        coord = AuxCoord(
            points=[0, 1],
            bounds=[[0, 1], [1, 2]],
            units="days since 1970-01-01",
            climatological=True,
        )
        self.assertTrue(coord.climatological)

    def test_create_no_bounds_no_set(self):
        with self.assertRaisesRegex(ValueError, "Cannot set.*no bounds exist"):
            AuxCoord(
                points=[0, 1],
                units="days since 1970-01-01",
                climatological=True,
            )

    def test_create_no_time_no_set(self):
        emsg = "Cannot set climatological .* valid time reference units.*"
        with self.assertRaisesRegex(TypeError, emsg):
            AuxCoord(
                points=[0, 1], bounds=[[0, 1], [1, 2]], climatological=True
            )

    def test_absent(self):
        coord = AuxCoord(points=[0, 1], bounds=[[0, 1], [1, 2]])
        self.assertFalse(coord.climatological)

    def test_absent_no_bounds_no_set(self):
        coord = AuxCoord(points=[0, 1], units="days since 1970-01-01")
        with self.assertRaisesRegex(ValueError, "Cannot set.*no bounds exist"):
            coord.climatological = True

    def test_absent_no_time_no_set(self):
        coord = AuxCoord(points=[0, 1], bounds=[[0, 1], [1, 2]])
        emsg = "Cannot set climatological .* valid time reference units.*"
        with self.assertRaisesRegex(TypeError, emsg):
            coord.climatological = True

    def test_absent_no_bounds_unset(self):
        coord = AuxCoord(points=[0, 1])
        coord.climatological = False
        self.assertFalse(coord.climatological)

    def test_bounds_set(self):
        coord = AuxCoord(
            points=[0, 1],
            bounds=[[0, 1], [1, 2]],
            units="days since 1970-01-01",
        )
        coord.climatological = True
        self.assertTrue(coord.climatological)

    def test_bounds_unset(self):
        coord = AuxCoord(
            points=[0, 1],
            bounds=[[0, 1], [1, 2]],
            units="days since 1970-01-01",
            climatological=True,
        )
        coord.climatological = False
        self.assertFalse(coord.climatological)

    def test_remove_bounds(self):
        coord = AuxCoord(
            points=[0, 1],
            bounds=[[0, 1], [1, 2]],
            units="days since 1970-01-01",
            climatological=True,
        )
        coord.bounds = None
        self.assertFalse(coord.climatological)

    def test_change_units(self):
        coord = AuxCoord(
            points=[0, 1],
            bounds=[[0, 1], [1, 2]],
            units="days since 1970-01-01",
            climatological=True,
        )
        self.assertTrue(coord.climatological)
        coord.units = "K"
        self.assertFalse(coord.climatological)


class Test___init____abstractmethod(tests.IrisTest):
    def test(self):
        emsg = (
            "Can't instantiate abstract class Coord with abstract"
            " methods __init__"
        )
        with self.assertRaisesRegex(TypeError, emsg):
            _ = Coord(points=[0, 1])


class Test_cube_dims(tests.IrisTest):
    def test(self):
        # Check that "coord.cube_dims(cube)" calls "cube.coord_dims(coord)".
        mock_dims_result = mock.sentinel.COORD_DIMS
        mock_dims_call = mock.Mock(return_value=mock_dims_result)
        mock_cube = mock.Mock(Cube, coord_dims=mock_dims_call)
        test_coord = AuxCoord([1], long_name="test_name")

        result = test_coord.cube_dims(mock_cube)
        self.assertEqual(result, mock_dims_result)
        self.assertEqual(
            mock_dims_call.call_args_list, [mock.call(test_coord)]
        )


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for the :class:`iris.coords.CellMethod`.
"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from iris.common import BaseMetadata
from iris.coords import CellMethod, AuxCoord


class Test(tests.IrisTest):
    def setUp(self):
        self.method = "mean"

    def _check(self, token, coord, default=False):
        result = CellMethod(self.method, coords=coord)
        token = token if not default else BaseMetadata.DEFAULT_NAME
        expected = "{}: {}".format(self.method, token)
        self.assertEqual(str(result), expected)

    def test_coord_standard_name(self):
        token = "air_temperature"
        coord = AuxCoord(1, standard_name=token)
        self._check(token, coord)

    def test_coord_long_name(self):
        token = "long_name"
        coord = AuxCoord(1, long_name=token)
        self._check(token, coord)

    def test_coord_long_name_default(self):
        token = "long name"  # includes space
        coord = AuxCoord(1, long_name=token)
        self._check(token, coord, default=True)

    def test_coord_var_name(self):
        token = "var_name"
        coord = AuxCoord(1, var_name=token)
        self._check(token, coord)

    def test_coord_var_name_fail(self):
        token = "var name"  # includes space
        emsg = "is not a valid NetCDF variable name"
        with self.assertRaisesRegex(ValueError, emsg):
            AuxCoord(1, var_name=token)

    def test_coord_stash(self):
        token = "stash"
        coord = AuxCoord(1, attributes=dict(STASH=token))
        self._check(token, coord, default=True)

    def test_coord_stash_default(self):
        token = "_stash"  # includes leading underscore
        coord = AuxCoord(1, attributes=dict(STASH=token))
        self._check(token, coord, default=True)

    def test_string(self):
        token = "air_temperature"
        result = CellMethod(self.method, coords=token)
        expected = "{}: {}".format(self.method, token)
        self.assertEqual(str(result), expected)

    def test_string_default(self):
        token = "air temperature"  # includes space
        result = CellMethod(self.method, coords=token)
        expected = "{}: unknown".format(self.method)
        self.assertEqual(str(result), expected)

    def test_mixture(self):
        token = "air_temperature"
        coord = AuxCoord(1, standard_name=token)
        result = CellMethod(self.method, coords=[coord, token])
        expected = "{}: {}, {}".format(self.method, token, token)
        self.assertEqual(str(result), expected)

    def test_mixture_default(self):
        token = "air temperature"  # includes space
        coord = AuxCoord(1, long_name=token)
        result = CellMethod(self.method, coords=[coord, token])
        expected = "{}: unknown, unknown".format(self.method)
        self.assertEqual(str(result), expected)


if __name__ == "__main__":
    tests.main()



import matplotlib.pyplot as plt
import numpy as np

import iris.palette


def main():
    a = np.linspace(0, 1, 256).reshape(1, -1)
    a = np.vstack((a, a))

    maps = sorted(iris.palette.CMAP_BREWER)
    nmaps = len(maps)

    fig = plt.figure(figsize=(7, 10))
    fig.subplots_adjust(top=0.99, bottom=0.01, left=0.2, right=0.99)
    for i, m in enumerate(maps):
        ax = plt.subplot(nmaps, 1, i + 1)
        plt.axis("off")
        plt.imshow(a, aspect="auto", cmap=plt.get_cmap(m), origin="lower")
        pos = list(ax.get_position().bounds)
        fig.text(
            pos[0] - 0.01, pos[1], m, fontsize=8, horizontalalignment="right"
        )

    plt.show()


if __name__ == "__main__":
    main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for the :class:`iris.coord_systems.AlbersEqualArea` class.

"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import cartopy.crs as ccrs
from iris.coord_systems import GeogCS, AlbersEqualArea


class Test_as_cartopy_crs(tests.IrisTest):
    def setUp(self):
        self.latitude_of_projection_origin = 0.0
        self.longitude_of_central_meridian = 0.0
        self.semi_major_axis = 6377563.396
        self.semi_minor_axis = 6356256.909
        self.false_easting = 0.0
        self.false_northing = 0.0
        self.standard_parallels = (-18.0, -36.0)
        self.ellipsoid = GeogCS(self.semi_major_axis, self.semi_minor_axis)
        self.aea_cs = AlbersEqualArea(
            self.latitude_of_projection_origin,
            self.longitude_of_central_meridian,
            self.false_easting,
            self.false_northing,
            self.standard_parallels,
            ellipsoid=self.ellipsoid,
        )

    def test_crs_creation(self):
        res = self.aea_cs.as_cartopy_crs()
        globe = ccrs.Globe(
            semimajor_axis=self.semi_major_axis,
            semiminor_axis=self.semi_minor_axis,
            ellipse=None,
        )
        expected = ccrs.AlbersEqualArea(
            self.longitude_of_central_meridian,
            self.latitude_of_projection_origin,
            self.false_easting,
            self.false_northing,
            self.standard_parallels,
            globe=globe,
        )
        self.assertEqual(res, expected)

    def test_fail_too_few_parallels(self):
        emsg = "parallels"
        with self.assertRaisesRegex(ValueError, emsg):
            AlbersEqualArea(standard_parallels=())

    def test_fail_too_many_parallels(self):
        emsg = "parallels"
        with self.assertRaisesRegex(ValueError, emsg):
            AlbersEqualArea(standard_parallels=(1, 2, 3))


class Test_as_cartopy_projection(tests.IrisTest):
    def setUp(self):
        self.latitude_of_projection_origin = 0.0
        self.longitude_of_central_meridian = 0.0
        self.semi_major_axis = 6377563.396
        self.semi_minor_axis = 6356256.909
        self.false_easting = 0.0
        self.false_northing = 0.0
        self.standard_parallels = (-18.0, -36.0)
        self.ellipsoid = GeogCS(self.semi_major_axis, self.semi_minor_axis)
        self.aea_cs = AlbersEqualArea(
            self.latitude_of_projection_origin,
            self.longitude_of_central_meridian,
            self.false_easting,
            self.false_northing,
            self.standard_parallels,
            ellipsoid=self.ellipsoid,
        )

    def test_projection_creation(self):
        res = self.aea_cs.as_cartopy_projection()
        globe = ccrs.Globe(
            semimajor_axis=self.semi_major_axis,
            semiminor_axis=self.semi_minor_axis,
            ellipse=None,
        )
        expected = ccrs.AlbersEqualArea(
            self.latitude_of_projection_origin,
            self.longitude_of_central_meridian,
            self.false_easting,
            self.false_northing,
            self.standard_parallels,
            globe=globe,
        )
        self.assertEqual(res, expected)


class Test_init_defaults(tests.IrisTest):
    def test_set_optional_args(self):
        # Check that setting optional arguments works as expected.
        crs = AlbersEqualArea(
            longitude_of_central_meridian=123,
            latitude_of_projection_origin=-17,
            false_easting=100,
            false_northing=-200,
            standard_parallels=(-37, 21.4),
        )

        self.assertEqualAndKind(crs.longitude_of_central_meridian, 123.0)
        self.assertEqualAndKind(crs.latitude_of_projection_origin, -17.0)
        self.assertEqualAndKind(crs.false_easting, 100.0)
        self.assertEqualAndKind(crs.false_northing, -200.0)
        self.assertEqual(len(crs.standard_parallels), 2)
        self.assertEqualAndKind(crs.standard_parallels[0], -37.0)
        self.assertEqualAndKind(crs.standard_parallels[1], 21.4)

    def _check_crs_defaults(self, crs):
        # Check for property defaults when no kwargs options were set.
        # NOTE: except ellipsoid, which is done elsewhere.
        self.assertEqualAndKind(crs.longitude_of_central_meridian, 0.0)
        self.assertEqualAndKind(crs.latitude_of_projection_origin, 0.0)
        self.assertEqualAndKind(crs.false_easting, 0.0)
        self.assertEqualAndKind(crs.false_northing, 0.0)
        self.assertEqual(len(crs.standard_parallels), 2)
        self.assertEqualAndKind(crs.standard_parallels[0], 20.0)
        self.assertEqualAndKind(crs.standard_parallels[1], 50.0)

    def test_no_optional_args(self):
        # Check expected defaults with no optional args.
        crs = AlbersEqualArea()
        self._check_crs_defaults(crs)

    def test_optional_args_None(self):
        # Check expected defaults with optional args=None.
        crs = AlbersEqualArea(
            longitude_of_central_meridian=None,
            latitude_of_projection_origin=None,
            standard_parallels=None,
            false_easting=None,
            false_northing=None,
        )
        self._check_crs_defaults(crs)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test function :func:`iris.fileformats._pyke_rules.compiled_krb.\
fc_rules_cf_fc.get_cf_bounds_var`.

"""

# import iris tests first so that some things can be initialised before
# importing anything else
import iris.tests as tests

from unittest import mock

from iris.fileformats._pyke_rules.compiled_krb.fc_rules_cf_fc import \
    get_cf_bounds_var, CF_ATTR_BOUNDS, CF_ATTR_CLIMATOLOGY


class TestGetCFBoundsVar(tests.IrisTest):
    # Tests to check that get_cf_bounds_var will return the bounds_var and
    # the correct climatological flag.
    def _generic_test(self, test_climatological_bounds=False):
        cf_coord_var = mock.MagicMock()

        cf_group_dict = {'TEST': mock.sentinel.bounds_var}
        if test_climatological_bounds:
            cf_coord_var.cf_group.climatology = cf_group_dict
            test_attr = CF_ATTR_CLIMATOLOGY
        else:
            cf_coord_var.cf_group.bounds = cf_group_dict
            test_attr = CF_ATTR_BOUNDS

        for attr in (CF_ATTR_BOUNDS, CF_ATTR_CLIMATOLOGY):
            attr_val = 'TEST' if attr == test_attr else None
            setattr(cf_coord_var, attr, attr_val)

        bounds_var, climatological = get_cf_bounds_var(cf_coord_var)
        self.assertIs(bounds_var, mock.sentinel.bounds_var)
        self.assertEqual(climatological, test_climatological_bounds)

    def test_bounds_normal(self):
        self._generic_test(test_climatological_bounds=False)

    def test_bounds_climatological(self):
        self._generic_test(test_climatological_bounds=True)


if __name__ == '__main__':
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the `iris.fileformats.pp.PPDataProxy` class."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from unittest import mock

from iris.fileformats.pp import PPDataProxy, SplittableInt


class Test_lbpack(tests.IrisTest):
    def test_lbpack_SplittableInt(self):
        lbpack = mock.Mock(spec_set=SplittableInt)
        proxy = PPDataProxy(None, None, None, None, None, lbpack, None, None)
        self.assertEqual(proxy.lbpack, lbpack)
        self.assertIs(proxy.lbpack, lbpack)

    def test_lbpack_raw(self):
        lbpack = 4321
        proxy = PPDataProxy(None, None, None, None, None, lbpack, None, None)
        self.assertEqual(proxy.lbpack, lbpack)
        self.assertIsNot(proxy.lbpack, lbpack)
        self.assertIsInstance(proxy.lbpack, SplittableInt)
        self.assertEqual(proxy.lbpack.n1, lbpack % 10)
        self.assertEqual(proxy.lbpack.n2, lbpack // 10 % 10)
        self.assertEqual(proxy.lbpack.n3, lbpack // 100 % 10)
        self.assertEqual(proxy.lbpack.n4, lbpack // 1000 % 10)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the `iris.fileformats.netcdf._load_cube` function."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from unittest import mock

import numpy as np

from iris.coords import DimCoord
import iris.fileformats.cf
from iris.fileformats.netcdf import _load_cube


class TestCoordAttributes(tests.IrisTest):
    @staticmethod
    def _patcher(engine, cf, cf_group):
        coordinates = []
        for coord in cf_group:
            engine.cube.add_aux_coord(coord)
            coordinates.append((coord, coord.name()))
        engine.cube_parts["coordinates"] = coordinates

    def setUp(self):
        this = "iris.fileformats.netcdf._assert_case_specific_facts"
        patch = mock.patch(this, side_effect=self._patcher)
        patch.start()
        self.addCleanup(patch.stop)
        self.engine = mock.Mock()
        self.filename = "DUMMY"
        self.flag_masks = mock.sentinel.flag_masks
        self.flag_meanings = mock.sentinel.flag_meanings
        self.flag_values = mock.sentinel.flag_values
        self.valid_range = mock.sentinel.valid_range
        self.valid_min = mock.sentinel.valid_min
        self.valid_max = mock.sentinel.valid_max

    def _make(self, names, attrs):
        coords = [DimCoord(i, long_name=name) for i, name in enumerate(names)]
        shape = (1,)

        cf_group = {}
        for name, cf_attrs in zip(names, attrs):
            cf_attrs_unused = mock.Mock(return_value=cf_attrs)
            cf_group[name] = mock.Mock(cf_attrs_unused=cf_attrs_unused)
        cf = mock.Mock(cf_group=cf_group)

        cf_data = mock.Mock(_FillValue=None)
        cf_data.chunking = mock.MagicMock(return_value=shape)
        cf_var = mock.MagicMock(
            spec=iris.fileformats.cf.CFVariable,
            dtype=np.dtype("i4"),
            cf_data=cf_data,
            cf_name="DUMMY_VAR",
            cf_group=coords,
            shape=shape,
        )
        return cf, cf_var

    def test_flag_pass_thru(self):
        items = [
            ("masks", "flag_masks", self.flag_masks),
            ("meanings", "flag_meanings", self.flag_meanings),
            ("values", "flag_values", self.flag_values),
        ]
        for name, attr, value in items:
            names = [name]
            attrs = [[(attr, value)]]
            cf, cf_var = self._make(names, attrs)
            cube = _load_cube(self.engine, cf, cf_var, self.filename)
            self.assertEqual(len(cube.coords(name)), 1)
            coord = cube.coord(name)
            self.assertEqual(len(coord.attributes), 1)
            self.assertEqual(list(coord.attributes.keys()), [attr])
            self.assertEqual(list(coord.attributes.values()), [value])

    def test_flag_pass_thru_multi(self):
        names = ["masks", "meanings", "values"]
        attrs = [
            [("flag_masks", self.flag_masks), ("wibble", "wibble")],
            [
                ("flag_meanings", self.flag_meanings),
                ("add_offset", "add_offset"),
            ],
            [("flag_values", self.flag_values)],
            [("valid_range", self.valid_range)],
            [("valid_min", self.valid_min)],
            [("valid_max", self.valid_max)],
        ]
        cf, cf_var = self._make(names, attrs)
        cube = _load_cube(self.engine, cf, cf_var, self.filename)
        self.assertEqual(len(cube.coords()), 3)
        self.assertEqual(set([c.name() for c in cube.coords()]), set(names))
        expected = [
            attrs[0],
            [attrs[1][0]],
            attrs[2],
            attrs[3],
            attrs[4],
            attrs[5],
        ]
        for name, expect in zip(names, expected):
            attributes = cube.coord(name).attributes
            self.assertEqual(set(attributes.items()), set(expect))


class TestCubeAttributes(tests.IrisTest):
    def setUp(self):
        this = "iris.fileformats.netcdf._assert_case_specific_facts"
        patch = mock.patch(this)
        patch.start()
        self.addCleanup(patch.stop)
        self.engine = mock.Mock()
        self.cf = None
        self.filename = "DUMMY"
        self.flag_masks = mock.sentinel.flag_masks
        self.flag_meanings = mock.sentinel.flag_meanings
        self.flag_values = mock.sentinel.flag_values
        self.valid_range = mock.sentinel.valid_range
        self.valid_min = mock.sentinel.valid_min
        self.valid_max = mock.sentinel.valid_max

    def _make(self, attrs):
        shape = (1,)
        cf_attrs_unused = mock.Mock(return_value=attrs)
        cf_data = mock.Mock(_FillValue=None)
        cf_data.chunking = mock.MagicMock(return_value=shape)
        cf_var = mock.MagicMock(
            spec=iris.fileformats.cf.CFVariable,
            dtype=np.dtype("i4"),
            cf_data=cf_data,
            cf_name="DUMMY_VAR",
            cf_group=mock.Mock(),
            cf_attrs_unused=cf_attrs_unused,
            shape=shape,
        )
        return cf_var

    def test_flag_pass_thru(self):
        attrs = [
            ("flag_masks", self.flag_masks),
            ("flag_meanings", self.flag_meanings),
            ("flag_values", self.flag_values),
        ]
        for key, value in attrs:
            cf_var = self._make([(key, value)])
            cube = _load_cube(self.engine, self.cf, cf_var, self.filename)
            self.assertEqual(len(cube.attributes), 1)
            self.assertEqual(list(cube.attributes.keys()), [key])
            self.assertEqual(list(cube.attributes.values()), [value])

    def test_flag_pass_thru_multi(self):
        attrs = [
            ("flag_masks", self.flag_masks),
            ("wibble", "wobble"),
            ("flag_meanings", self.flag_meanings),
            ("add_offset", "add_offset"),
            ("flag_values", self.flag_values),
            ("standard_name", "air_temperature"),
            ("valid_range", self.valid_range),
            ("valid_min", self.valid_min),
            ("valid_max", self.valid_max),
        ]

        # Expect everything from above to be returned except those
        # corresponding to exclude_ind.
        expected = set([attrs[ind] for ind in [0, 1, 2, 4, 6, 7, 8]])
        cf_var = self._make(attrs)
        cube = _load_cube(self.engine, self.cf, cf_var, self.filename)
        self.assertEqual(len(cube.attributes), len(expected))
        self.assertEqual(set(cube.attributes.items()), expected)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Test function :func:`iris.util.promote_aux_coord_to_dim_coord`."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import unittest

import iris
import iris.tests.stock as stock
from iris.util import promote_aux_coord_to_dim_coord


class Test(tests.IrisTest):
    def test_dimension_already_has_dimcoord(self):
        cube_a = stock.hybrid_height()
        cube_b = cube_a.copy()
        promote_aux_coord_to_dim_coord(cube_b, "model_level_number")
        self.assertEqual(
            cube_b.dim_coords, (cube_a.coord("model_level_number"),)
        )

    def test_old_dim_coord_is_now_aux_coord(self):
        cube_a = stock.hybrid_height()
        cube_b = cube_a.copy()
        promote_aux_coord_to_dim_coord(cube_b, "model_level_number")
        self.assertTrue(cube_a.coord("level_height") in cube_b.aux_coords)

    @tests.skip_data
    def test_argument_is_coord_instance(self):
        cube_a = stock.realistic_4d()
        cube_b = cube_a.copy()
        promote_aux_coord_to_dim_coord(cube_b, cube_b.coord("level_height"))
        self.assertEqual(
            cube_b.dim_coords,
            (
                cube_a.coord("time"),
                cube_a.coord("level_height"),
                cube_a.coord("grid_latitude"),
                cube_a.coord("grid_longitude"),
            ),
        )

    @tests.skip_data
    def test_dimension_is_anonymous(self):
        cube_a = stock.realistic_4d()
        cube_b = cube_a.copy()
        cube_b.remove_coord("model_level_number")
        promote_aux_coord_to_dim_coord(cube_b, "level_height")
        self.assertEqual(
            cube_b.dim_coords,
            (
                cube_a.coord("time"),
                cube_a.coord("level_height"),
                cube_a.coord("grid_latitude"),
                cube_a.coord("grid_longitude"),
            ),
        )

    def test_already_a_dim_coord(self):
        cube_a = stock.simple_2d_w_multidim_and_scalars()
        cube_b = cube_a.copy()
        promote_aux_coord_to_dim_coord(cube_b, "dim1")
        self.assertEqual(cube_a, cube_b)

    def test_coord_of_that_name_does_not_exist(self):
        cube_a = stock.simple_2d_w_multidim_and_scalars()
        with self.assertRaises(iris.exceptions.CoordinateNotFoundError):
            promote_aux_coord_to_dim_coord(cube_a, "wibble")

    def test_coord_does_not_exist(self):
        cube_a = stock.simple_2d_w_multidim_and_scalars()
        coord = cube_a.coord("dim1").copy()
        coord.rename("new")
        with self.assertRaises(ValueError):
            promote_aux_coord_to_dim_coord(cube_a, coord)

    def test_argument_is_wrong_type(self):
        cube_a = stock.simple_1d()
        with self.assertRaises(TypeError):
            promote_aux_coord_to_dim_coord(cube_a, 0.0)

    def test_trying_to_promote_a_multidim_coord(self):
        cube_a = stock.simple_2d_w_multidim_coords()
        with self.assertRaises(ValueError):
            promote_aux_coord_to_dim_coord(cube_a, "bar")

    def test_trying_to_promote_a_scalar_coord(self):
        cube_a = stock.simple_2d_w_multidim_and_scalars()
        with self.assertRaises(ValueError):
            promote_aux_coord_to_dim_coord(cube_a, "an_other")

    def test_trying_to_promote_a_nonmonotonic_coord(self):
        cube_a = stock.hybrid_height()
        with self.assertRaises(ValueError):
            promote_aux_coord_to_dim_coord(cube_a, "surface_altitude")


if __name__ == "__main__":
    unittest.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Integration tests for :func:`iris.fileformats.rules.load_cubes`."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import iris
from iris.fileformats import pp
from iris.fileformats.pp_load_rules import convert
from iris.fileformats.rules import load_cubes


class Test(tests.IrisTest):
    @tests.skip_data
    def test_pp_with_stash_constraint(self):
        filenames = [tests.get_data_path(("PP", "globClim1", "dec_subset.pp"))]
        stcon = iris.AttributeConstraint(STASH="m01s00i004")
        pp_constraints = pp._convert_constraints(stcon)
        pp_loader = iris.fileformats.rules.Loader(pp.load, {}, convert)
        cubes = list(load_cubes(filenames, None, pp_loader, pp_constraints))
        self.assertEqual(len(cubes), 38)

    @tests.skip_data
    def test_pp_with_stash_constraints(self):
        filenames = [tests.get_data_path(("PP", "globClim1", "dec_subset.pp"))]
        stcon1 = iris.AttributeConstraint(STASH="m01s00i004")
        stcon2 = iris.AttributeConstraint(STASH="m01s00i010")
        pp_constraints = pp._convert_constraints([stcon1, stcon2])
        pp_loader = iris.fileformats.rules.Loader(pp.load, {}, convert)
        cubes = list(load_cubes(filenames, None, pp_loader, pp_constraints))
        self.assertEqual(len(cubes), 76)

    @tests.skip_data
    def test_pp_no_constraint(self):
        filenames = [tests.get_data_path(("PP", "globClim1", "dec_subset.pp"))]
        pp_constraints = pp._convert_constraints(None)
        pp_loader = iris.fileformats.rules.Loader(pp.load, {}, convert)
        cubes = list(load_cubes(filenames, None, pp_loader, pp_constraints))
        self.assertEqual(len(cubes), 152)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Integration tests for :class:`iris.cube.Cube`."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import numpy as np

import iris
from iris.analysis import MEAN
from iris.cube import Cube

from iris._lazy_data import is_lazy_data, as_lazy_data


class Test_aggregated_by(tests.IrisTest):
    @tests.skip_data
    def test_agg_by_aux_coord(self):
        problem_test_file = tests.get_data_path(
            ("NetCDF", "testing", "small_theta_colpex.nc")
        )
        cube = iris.load_cube(problem_test_file, "air_potential_temperature")

        # Test aggregating by aux coord, notably the `forecast_period` aux
        # coord on `cube`, whose `_points` attribute is a lazy array.
        # This test then ensures that aggregating using `points` instead is
        # successful.

        # First confirm we've got a lazy array.
        # NB. This checks the merge process in `load_cube()` hasn't
        # triggered the load of the coordinate's data.
        forecast_period_coord = cube.coord("forecast_period")

        self.assertTrue(is_lazy_data(forecast_period_coord.core_points()))

        # Now confirm we can aggregate along this coord.
        res_cube = cube.aggregated_by("forecast_period", MEAN)
        res_cell_methods = res_cube.cell_methods[0]
        self.assertEqual(res_cell_methods.coord_names, ("forecast_period",))
        self.assertEqual(res_cell_methods.method, "mean")


class TestDataFillValue(tests.IrisTest):
    def test_real(self):
        data = np.ma.masked_array([1, 2, 3], [0, 1, 0], fill_value=10)
        cube = Cube(data)
        cube.data.fill_value = 20
        self.assertEqual(cube.data.fill_value, 20)

    def test_lazy(self):
        data = np.ma.masked_array([1, 2, 3], [0, 1, 0], fill_value=10)
        data = as_lazy_data(data)
        cube = Cube(data)
        cube.data.fill_value = 20
        self.assertEqual(cube.data.fill_value, 20)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the `iris.fileformats.netcdf._get_cf_var_data` function."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from unittest import mock

from dask.array import Array as dask_array
import numpy as np

from iris._lazy_data import _optimum_chunksize
import iris.fileformats.cf
from iris.fileformats.netcdf import _get_cf_var_data


class Test__get_cf_var_data(tests.IrisTest):
    def setUp(self):
        self.filename = "DUMMY"
        self.shape = (300000, 240, 200)
        self.expected_chunks = _optimum_chunksize(self.shape, self.shape)

    def _make(self, chunksizes):
        cf_data = mock.Mock(_FillValue=None)
        cf_data.chunking = mock.MagicMock(return_value=chunksizes)
        cf_var = mock.MagicMock(
            spec=iris.fileformats.cf.CFVariable,
            dtype=np.dtype("i4"),
            cf_data=cf_data,
            cf_name="DUMMY_VAR",
            shape=self.shape,
        )
        return cf_var

    def test_cf_data_type(self):
        chunks = [1, 12, 100]
        cf_var = self._make(chunks)
        lazy_data = _get_cf_var_data(cf_var, self.filename)
        self.assertIsInstance(lazy_data, dask_array)

    def test_cf_data_chunks(self):
        chunks = [2500, 240, 200]
        cf_var = self._make(chunks)
        lazy_data = _get_cf_var_data(cf_var, self.filename)
        lazy_data_chunks = [c[0] for c in lazy_data.chunks]
        expected_chunks = _optimum_chunksize(chunks, self.shape)
        self.assertArrayEqual(lazy_data_chunks, expected_chunks)

    def test_cf_data_no_chunks(self):
        # No chunks means chunks are calculated from the array's shape by
        # `iris._lazy_data._optimum_chunksize()`.
        chunks = None
        cf_var = self._make(chunks)
        lazy_data = _get_cf_var_data(cf_var, self.filename)
        lazy_data_chunks = [c[0] for c in lazy_data.chunks]
        self.assertArrayEqual(lazy_data_chunks, self.expected_chunks)

    def test_cf_data_contiguous(self):
        # Chunks 'contiguous' is equivalent to no chunks.
        chunks = "contiguous"
        cf_var = self._make(chunks)
        lazy_data = _get_cf_var_data(cf_var, self.filename)
        lazy_data_chunks = [c[0] for c in lazy_data.chunks]
        self.assertArrayEqual(lazy_data_chunks, self.expected_chunks)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for
:func:`iris.fileformats.pp_load_rules._collapse_degenerate_points_and_bounds`.

"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import numpy as np

from iris.fileformats.pp_load_rules import (
    _collapse_degenerate_points_and_bounds,
)


class Test(tests.IrisTest):
    def test_scalar(self):
        array = np.array(1)
        points, bounds = _collapse_degenerate_points_and_bounds(array)
        self.assertArrayEqual(points, array)
        self.assertIsNone(bounds)

    def test_1d_nochange(self):
        array = np.array([1, 1, 3])
        result, _ = _collapse_degenerate_points_and_bounds(array)
        self.assertArrayEqual(result, array)

    def test_1d_collapse(self):
        array = np.array([1, 1, 1])
        result, _ = _collapse_degenerate_points_and_bounds(array)
        self.assertArrayEqual(result, np.array([1]))

    def test_2d_nochange(self):
        array = np.array([[1, 2, 3], [4, 5, 6]])
        result, _ = _collapse_degenerate_points_and_bounds(array)
        self.assertArrayEqual(result, array)

    def test_2d_collapse_dim0(self):
        array = np.array([[1, 2, 3], [1, 2, 3]])
        result, _ = _collapse_degenerate_points_and_bounds(array)
        self.assertArrayEqual(result, np.array([[1, 2, 3]]))

    def test_2d_collapse_dim1(self):
        array = np.array([[1, 1, 1], [2, 2, 2]])
        result, _ = _collapse_degenerate_points_and_bounds(array)
        self.assertArrayEqual(result, np.array([[1], [2]]))

    def test_2d_collapse_both(self):
        array = np.array([[3, 3, 3], [3, 3, 3]])
        result, _ = _collapse_degenerate_points_and_bounds(array)
        self.assertArrayEqual(result, np.array([[3]]))

    def test_3d(self):
        array = np.array([[[3, 3, 3], [4, 4, 4]], [[3, 3, 3], [4, 4, 4]]])
        result, _ = _collapse_degenerate_points_and_bounds(array)
        self.assertArrayEqual(result, np.array([[[3], [4]]]))

    def test_multiple_odd_dims(self):
        # Test to ensure multiple collapsed dimensions don't interfere.
        # make a 5-D array where dimensions 0, 2 and 3 are degenerate.
        array = np.arange(3 ** 5).reshape([3] * 5)
        array[1:] = array[0:1]
        array[:, :, 1:] = array[:, :, 0:1]
        array[:, :, :, 1:] = array[:, :, :, 0:1]
        result, _ = _collapse_degenerate_points_and_bounds(array)
        self.assertEqual(array.shape, (3, 3, 3, 3, 3))
        self.assertEqual(result.shape, (1, 3, 1, 1, 3))
        self.assertTrue(np.all(result == array[0:1, :, 0:1, 0:1, :]))

    def test_bounds_collapse(self):
        points = np.array([1, 1, 1])
        bounds = np.array([[0, 1], [0, 1], [0, 1]])
        result_pts, result_bds = _collapse_degenerate_points_and_bounds(
            points, bounds
        )
        self.assertArrayEqual(result_pts, np.array([1]))
        self.assertArrayEqual(result_bds, np.array([[0, 1]]))

    def test_bounds_no_collapse(self):
        points = np.array([1, 1, 1])
        bounds = np.array([[0, 1], [0, 1], [0, 2]])
        result_pts, result_bds = _collapse_degenerate_points_and_bounds(
            points, bounds
        )
        self.assertArrayEqual(result_pts, points)
        self.assertArrayEqual(result_bds, bounds)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the `iris.io.expand_filespecs` function."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import os
import tempfile
import shutil
import textwrap

import iris.io as iio


class TestExpandFilespecs(tests.IrisTest):
    def setUp(self):
        tests.IrisTest.setUp(self)
        self.tmpdir = os.path.realpath(tempfile.mkdtemp())
        self.fnames = ["a.foo", "b.txt"]
        for fname in self.fnames:
            with open(os.path.join(self.tmpdir, fname), "w") as fh:
                fh.write("anything")

    def tearDown(self):
        shutil.rmtree(self.tmpdir)

    def test_absolute_path(self):
        result = iio.expand_filespecs([os.path.join(self.tmpdir, "*")])
        expected = [os.path.join(self.tmpdir, fname) for fname in self.fnames]
        self.assertEqual(result, expected)

    def test_double_slash(self):
        product = iio.expand_filespecs(["//" + os.path.join(self.tmpdir, "*")])
        predicted = [os.path.join(self.tmpdir, fname) for fname in self.fnames]
        self.assertEqual(product, predicted)

    def test_relative_path(self):
        cwd = os.getcwd()
        try:
            os.chdir(self.tmpdir)
            item_out = iio.expand_filespecs(["*"])
            item_in = [
                os.path.join(self.tmpdir, fname) for fname in self.fnames
            ]
            self.assertEqual(item_out, item_in)
        finally:
            os.chdir(cwd)

    def test_return_order(self):
        # It is really quite important what order we return the
        # files. They should be in the order that was provided,
        # so that we can control the order of load (for instance,
        # this can be used with PP files to ensure that there is
        # a surface reference).
        patterns = [
            os.path.join(self.tmpdir, "a.*"),
            os.path.join(self.tmpdir, "b.*"),
        ]
        expected = [
            os.path.join(self.tmpdir, fname) for fname in ["a.foo", "b.txt"]
        ]
        result = iio.expand_filespecs(patterns)
        self.assertEqual(result, expected)
        result = iio.expand_filespecs(patterns[::-1])
        self.assertEqual(result, expected[::-1])

    def test_no_files_found(self):
        msg = r"\/no_exist.txt\" didn\'t match any files"
        with self.assertRaisesRegex(IOError, msg):
            iio.expand_filespecs([os.path.join(self.tmpdir, "no_exist.txt")])

    def test_files_and_none(self):
        with self.assertRaises(IOError) as err:
            iio.expand_filespecs(
                [
                    os.path.join(self.tmpdir, "does_not_exist.txt"),
                    os.path.join(self.tmpdir, "*"),
                ]
            )
        expected = (
            textwrap.dedent(
                """
            One or more of the files specified did not exist:
                * "{0}/does_not_exist.txt" didn\'t match any files
                - "{0}/*" matched 2 file(s)
            """
            )
            .strip()
            .format(self.tmpdir)
        )

        self.assertStringEqual(str(err.exception), expected)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test function :func:`iris.fileformats._pyke_rules.compiled_krb.\
fc_rules_cf_fc.has_supported_mercator_parameters`.

"""

import warnings

# import iris tests first so that some things can be initialised before
# importing anything else
import iris.tests as tests

from unittest import mock

from iris.fileformats._pyke_rules.compiled_krb.fc_rules_cf_fc import \
    has_supported_mercator_parameters


def _engine(cf_grid_var, cf_name):
    cf_group = {cf_name: cf_grid_var}
    cf_var = mock.Mock(cf_group=cf_group)
    return mock.Mock(cf_var=cf_var)


class TestHasSupportedMercatorParameters(tests.IrisTest):

    def test_valid(self):
        cf_name = 'mercator'
        cf_grid_var = mock.Mock(
            spec=[],
            longitude_of_projection_origin=-90,
            false_easting=0,
            false_northing=0,
            scale_factor_at_projection_origin=1,
            semi_major_axis=6377563.396,
            semi_minor_axis=6356256.909)
        engine = _engine(cf_grid_var, cf_name)

        is_valid = has_supported_mercator_parameters(engine, cf_name)

        self.assertTrue(is_valid)

    def test_invalid_scale_factor(self):
        # Iris does not yet support scale factors other than one for
        # Mercator projections
        cf_name = 'mercator'
        cf_grid_var = mock.Mock(
            spec=[],
            longitude_of_projection_origin=0,
            false_easting=0,
            false_northing=0,
            scale_factor_at_projection_origin=0.9,
            semi_major_axis=6377563.396,
            semi_minor_axis=6356256.909)
        engine = _engine(cf_grid_var, cf_name)

        with warnings.catch_warnings(record=True) as warns:
            warnings.simplefilter("always")
            is_valid = has_supported_mercator_parameters(engine, cf_name)

        self.assertFalse(is_valid)
        self.assertEqual(len(warns), 1)
        self.assertRegex(str(warns[0]), 'Scale factor')

    def test_invalid_standard_parallel(self):
        # Iris does not yet support standard parallels other than zero for
        # Mercator projections
        cf_name = 'mercator'
        cf_grid_var = mock.Mock(
            spec=[],
            longitude_of_projection_origin=0,
            false_easting=0,
            false_northing=0,
            standard_parallel=30,
            semi_major_axis=6377563.396,
            semi_minor_axis=6356256.909)
        engine = _engine(cf_grid_var, cf_name)

        with warnings.catch_warnings(record=True) as warns:
            warnings.simplefilter("always")
            is_valid = has_supported_mercator_parameters(engine, cf_name)

        self.assertFalse(is_valid)
        self.assertEqual(len(warns), 1)
        self.assertRegex(str(warns[0]), 'Standard parallel')

    def test_invalid_false_easting(self):
        # Iris does not yet support false eastings other than zero for
        # Mercator projections
        cf_name = 'mercator'
        cf_grid_var = mock.Mock(
            spec=[],
            longitude_of_projection_origin=0,
            false_easting=100,
            false_northing=0,
            scale_factor_at_projection_origin=1,
            semi_major_axis=6377563.396,
            semi_minor_axis=6356256.909)
        engine = _engine(cf_grid_var, cf_name)

        with warnings.catch_warnings(record=True) as warns:
            warnings.simplefilter("always")
            is_valid = has_supported_mercator_parameters(engine, cf_name)

        self.assertFalse(is_valid)
        self.assertEqual(len(warns), 1)
        self.assertRegex(str(warns[0]), 'False easting')

    def test_invalid_false_northing(self):
        # Iris does not yet support false northings other than zero for
        # Mercator projections
        cf_name = 'mercator'
        cf_grid_var = mock.Mock(
            spec=[],
            longitude_of_projection_origin=0,
            false_easting=0,
            false_northing=100,
            scale_factor_at_projection_origin=1,
            semi_major_axis=6377563.396,
            semi_minor_axis=6356256.909)
        engine = _engine(cf_grid_var, cf_name)

        with warnings.catch_warnings(record=True) as warns:
            warnings.simplefilter("always")
            is_valid = has_supported_mercator_parameters(engine, cf_name)

        self.assertFalse(is_valid)
        self.assertEqual(len(warns), 1)
        self.assertRegex(str(warns[0]), 'False northing')


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test plots with two dimensional coordinates.

"""

# import iris tests first so that some things can be initialised before
# importing anything else
import iris.tests as tests

import cartopy.crs as ccrs
import matplotlib.pyplot as plt
import numpy as np

import iris
from iris.analysis.cartography import unrotate_pole
from iris.cube import Cube
from iris.coords import AuxCoord


# Run tests in no graphics mode if matplotlib is not available.
if tests.MPL_AVAILABLE:
    import iris.quickplot as qplt


@tests.skip_data
def simple_cube_w_2d_coords():
    path = tests.get_data_path(("NetCDF", "ORCA2", "votemper.nc"))
    cube = iris.load_cube(path)
    return cube


@tests.skip_plot
@tests.skip_data
class Test(tests.GraphicsTest):
    def test_2d_coord_bounds_platecarree(self):
        # To avoid a problem with Cartopy smearing the data where the
        # longitude wraps, we set the central_longitude
        cube = simple_cube_w_2d_coords()[0, 0]
        ax = plt.axes(projection=ccrs.PlateCarree(central_longitude=180))
        qplt.pcolormesh(cube)
        ax.coastlines(resolution="110m", color="red")
        self.check_graphic()

    def test_2d_coord_bounds_northpolarstereo(self):
        cube = simple_cube_w_2d_coords()[0, 0]
        ax = plt.axes(projection=ccrs.NorthPolarStereo())
        qplt.pcolormesh(cube)
        ax.coastlines(resolution="110m", color="red")
        self.check_graphic()


@tests.skip_plot
class Test2dContour(tests.GraphicsTest):
    def test_2d_coords_contour(self):
        ny, nx = 4, 6
        x1 = np.linspace(-20, 70, nx)
        y1 = np.linspace(10, 60, ny)
        data = np.zeros((ny, nx))
        data.flat[:] = np.arange(nx * ny) % 7
        cube = Cube(data, long_name="Odd data")
        x2, y2 = np.meshgrid(x1, y1)
        true_lons, true_lats = unrotate_pole(x2, y2, -130.0, 77.0)
        co_x = AuxCoord(true_lons, standard_name="longitude", units="degrees")
        co_y = AuxCoord(true_lats, standard_name="latitude", units="degrees")
        cube.add_aux_coord(co_y, (0, 1))
        cube.add_aux_coord(co_x, (0, 1))
        ax = plt.axes(projection=ccrs.PlateCarree())
        qplt.contourf(cube)
        ax.coastlines(resolution="110m", color="red")
        ax.gridlines(draw_labels=True)
        ax.set_extent((0, 180, 0, 90))
        self.check_graphic()


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
High-level plotting extensions to :mod:`iris.plot`.

These routines work much like their :mod:`iris.plot` counterparts, but they
automatically add a plot title, axis titles, and a colour bar when appropriate.

See also: :ref:`matplotlib <matplotlib:users-guide-index>`.

"""

from packaging import version

import cf_units
from matplotlib import __version__ as _mpl_version
import matplotlib.pyplot as plt

import iris.config
import iris.coords
import iris.plot as iplt


def _use_symbol(units):
    # For non-time units use the shortest unit representation.
    # E.g. prefer 'K' over 'kelvin', but not '0.0174532925199433 rad'
    # over 'degrees'
    return (
        not units.is_time()
        and not units.is_time_reference()
        and len(units.symbol) < len(str(units))
    )


def _title(cube_or_coord, with_units):
    if cube_or_coord is None or isinstance(cube_or_coord, int):
        title = ""
    else:
        title = cube_or_coord.name().replace("_", " ").capitalize()
        units = cube_or_coord.units
        if with_units and not (
            units.is_unknown()
            or units.is_no_unit()
            or units == cf_units.Unit("1")
        ):

            if _use_symbol(units):
                units = units.symbol
            elif units.is_time_reference():
                # iris.plot uses matplotlib.dates.date2num, which is fixed to the below unit.
                if version.parse(_mpl_version) >= version.parse("3.3"):
                    days_since = "1970-01-01"
                else:
                    days_since = "0001-01-01"
                units = "days since {}".format(days_since)
            title += " / {}".format(units)

    return title


def _label(cube, mode, result=None, ndims=2, coords=None, axes=None):
    """Puts labels on the current plot using the given cube."""

    if axes is None:
        axes = plt.gca()

    axes.set_title(_title(cube, with_units=False))

    if result is not None:
        draw_edges = mode == iris.coords.POINT_MODE
        bar = plt.colorbar(
            result, orientation="horizontal", drawedges=draw_edges
        )
        has_known_units = not (
            cube.units.is_unknown() or cube.units.is_no_unit()
        )
        if has_known_units and cube.units != cf_units.Unit("1"):
            # Use shortest unit representation for anything other than time
            if _use_symbol(cube.units):
                bar.set_label(cube.units.symbol)
            else:
                bar.set_label(cube.units)
        # Remove the tick which is put on the colorbar by default.
        bar.ax.tick_params(length=0)

    if coords is None:
        plot_defn = iplt._get_plot_defn(cube, mode, ndims)
    else:
        plot_defn = iplt._get_plot_defn_custom_coords_picked(
            cube, coords, mode, ndims=ndims
        )

    if ndims == 2:
        if not iplt._can_draw_map(plot_defn.coords):
            axes.set_ylabel(_title(plot_defn.coords[0], with_units=True))
            axes.set_xlabel(_title(plot_defn.coords[1], with_units=True))
    elif ndims == 1:
        axes.set_xlabel(_title(plot_defn.coords[0], with_units=True))
        axes.set_ylabel(_title(cube, with_units=True))
    else:
        msg = (
            "Unexpected number of dimensions ({}) given to "
            "_label.".format(ndims)
        )
        raise ValueError(msg)


def _label_with_bounds(cube, result=None, ndims=2, coords=None, axes=None):
    _label(cube, iris.coords.BOUND_MODE, result, ndims, coords, axes)


def _label_with_points(cube, result=None, ndims=2, coords=None, axes=None):
    _label(cube, iris.coords.POINT_MODE, result, ndims, coords, axes)


def _get_titles(u_object, v_object):
    if u_object is None:
        u_object = iplt._u_object_from_v_object(v_object)
    xunits = u_object is not None and not u_object.units.is_time_reference()
    yunits = not v_object.units.is_time_reference()
    xlabel = _title(u_object, with_units=xunits)
    ylabel = _title(v_object, with_units=yunits)
    title = ""
    if u_object is None:
        title = _title(v_object, with_units=False)
    elif isinstance(u_object, iris.cube.Cube) and not isinstance(
        v_object, iris.cube.Cube
    ):
        title = _title(u_object, with_units=False)
    elif isinstance(v_object, iris.cube.Cube) and not isinstance(
        u_object, iris.cube.Cube
    ):
        title = _title(v_object, with_units=False)
    return xlabel, ylabel, title


def _label_1d_plot(*args, **kwargs):
    if len(args) > 1 and isinstance(
        args[1], (iris.cube.Cube, iris.coords.Coord)
    ):
        xlabel, ylabel, title = _get_titles(*args[:2])
    else:
        xlabel, ylabel, title = _get_titles(None, args[0])

    axes = kwargs.pop("axes", None)

    if len(kwargs) != 0:
        msg = "Unexpected kwargs {} given to _label_1d_plot".format(
            kwargs.keys()
        )
        raise ValueError(msg)

    if axes is None:
        axes = plt.gca()

    axes.set_title(title)
    axes.set_xlabel(xlabel)
    axes.set_ylabel(ylabel)


def contour(cube, *args, **kwargs):
    """
    Draws contour lines on a labelled plot based on the given Cube.

    With the basic call signature, contour "level" values are chosen
    automatically::

        contour(cube)

    Supply a number to use *N* automatically chosen levels::

        contour(cube, N)

    Supply a sequence *V* to use explicitly defined levels::

        contour(cube, V)

    See :func:`iris.plot.contour` for details of valid keyword arguments.

    """
    coords = kwargs.get("coords")
    axes = kwargs.get("axes")
    result = iplt.contour(cube, *args, **kwargs)
    _label_with_points(cube, coords=coords, axes=axes)
    return result


def contourf(cube, *args, **kwargs):
    """
    Draws filled contours on a labelled plot based on the given Cube.

    With the basic call signature, contour "level" values are chosen
    automatically::

        contour(cube)

    Supply a number to use *N* automatically chosen levels::

        contour(cube, N)

    Supply a sequence *V* to use explicitly defined levels::

        contour(cube, V)

    See :func:`iris.plot.contourf` for details of valid keyword arguments.

    """
    coords = kwargs.get("coords")
    axes = kwargs.get("axes")
    result = iplt.contourf(cube, *args, **kwargs)
    _label_with_points(cube, result, coords=coords, axes=axes)
    return result


def outline(cube, coords=None, color="k", linewidth=None, axes=None):
    """
    Draws cell outlines on a labelled plot based on the given Cube.

    Kwargs:

    * coords: list of :class:`~iris.coords.Coord` objects or coordinate names
        Use the given coordinates as the axes for the plot. The order of the
        given coordinates indicates which axis to use for each, where the first
        element is the horizontal axis of the plot and the second element is
        the vertical axis of the plot.

    * color: None or mpl color
        The color of the cell outlines. If None, the matplotlibrc setting
        patch.edgecolor is used by default.

    * linewidth: None or number
        The width of the lines showing the cell outlines. If None, the default
        width in patch.linewidth in matplotlibrc is used.

    """
    result = iplt.outline(
        cube, color=color, linewidth=linewidth, coords=coords, axes=axes
    )

    _label_with_bounds(cube, coords=coords, axes=axes)
    return result


def pcolor(cube, *args, **kwargs):
    """
    Draws a labelled pseudocolor plot based on the given Cube.

    See :func:`iris.plot.pcolor` for details of valid keyword arguments.

    """
    coords = kwargs.get("coords")
    axes = kwargs.get("axes")
    result = iplt.pcolor(cube, *args, **kwargs)
    _label_with_bounds(cube, result, coords=coords, axes=axes)
    return result


def pcolormesh(cube, *args, **kwargs):
    """
    Draws a labelled pseudocolour plot based on the given Cube.

    See :func:`iris.plot.pcolormesh` for details of valid keyword arguments.

    """
    coords = kwargs.get("coords")
    axes = kwargs.get("axes")
    result = iplt.pcolormesh(cube, *args, **kwargs)
    _label_with_bounds(cube, result, coords=coords, axes=axes)
    return result


def points(cube, *args, **kwargs):
    """
    Draws sample point positions on a labelled plot based on the given Cube.

    See :func:`iris.plot.points` for details of valid keyword arguments.

    """
    coords = kwargs.get("coords")
    axes = kwargs.get("axes")
    result = iplt.points(cube, *args, **kwargs)
    _label_with_points(cube, coords=coords, axes=axes)
    return result


def plot(*args, **kwargs):
    """
    Draws a labelled line plot based on the given cube(s) or
    coordinate(s).

    See :func:`iris.plot.plot` for details of valid arguments and
    keyword arguments.

    """
    axes = kwargs.get("axes")
    result = iplt.plot(*args, **kwargs)
    _label_1d_plot(*args, axes=axes)
    return result


def scatter(x, y, *args, **kwargs):
    """
    Draws a labelled scatter plot based on the given cubes or
    coordinates.

    See :func:`iris.plot.scatter` for details of valid arguments and
    keyword arguments.

    """
    axes = kwargs.get("axes")
    result = iplt.scatter(x, y, *args, **kwargs)
    _label_1d_plot(x, y, axes=axes)
    return result


# Provide a convenience show method from pyplot.
show = plt.show



import matplotlib.pyplot as plt
import iris
import iris.quickplot as qplt


# Load the data for a single value of model level number.
fname = iris.sample_data_path("hybrid_height.nc")
temperature_cube = iris.load_cube(fname, iris.Constraint(model_level_number=1))

# Draw the block plot.
qplt.pcolormesh(temperature_cube)

plt.show()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.

"""
Definitions of how Iris objects should be represented.

"""

from html import escape
import re


class CubeRepresentation:
    """
    Produce representations of a :class:`~iris.cube.Cube`.

    This includes:

    * ``_html_repr_``: a representation of the cube as an html object,
      available in Jupyter notebooks. Specifically, this is presented as an
      html table.

    """

    _template = """
<style>
  a.iris {{
      text-decoration: none !important;
  }}
  table.iris {{
      white-space: pre;
      border: 1px solid;
      border-color: #9c9c9c;
      font-family: monaco, monospace;
  }}
  th.iris {{
      background: #303f3f;
      color: #e0e0e0;
      border-left: 1px solid;
      border-color: #9c9c9c;
      font-size: 1.05em;
      min-width: 50px;
      max-width: 125px;
  }}
  tr.iris :first-child {{
      border-right: 1px solid #9c9c9c !important;
  }}
  td.iris-title {{
      background: #d5dcdf;
      border-top: 1px solid #9c9c9c;
      font-weight: bold;
  }}
  .iris-word-cell {{
      text-align: left !important;
      white-space: pre;
  }}
  .iris-subheading-cell {{
      padding-left: 2em !important;
  }}
  .iris-inclusion-cell {{
      padding-right: 1em !important;
  }}
  .iris-panel-body {{
      padding-top: 0px;
  }}
  .iris-panel-title {{
      padding-left: 3em;
  }}
  .iris-panel-title {{
      margin-top: 7px;
  }}
</style>
<table class="iris" id="{id}">
    {header}
    {shape}
    {content}
</table>
        """

    def __init__(self, cube):
        self.cube = cube
        self.cube_id = id(self.cube)
        self.cube_str = escape(str(self.cube))

        self.str_headings = {
            "Dimension coordinates:": None,
            "Auxiliary coordinates:": None,
            "Derived coordinates:": None,
            "Cell measures:": None,
            "Ancillary variables:": None,
            "Scalar coordinates:": None,
            "Scalar cell measures:": None,
            "Attributes:": None,
            "Cell methods:": None,
        }
        self.dim_desc_coords = [
            "Dimension coordinates:",
            "Auxiliary coordinates:",
            "Derived coordinates:",
            "Cell measures:",
            "Ancillary variables:",
        ]

        self.two_cell_headers = ["Scalar coordinates:", "Attributes:"]

        # Important content that summarises a cube is defined here.
        self.shapes = self.cube.shape
        self.scalar_cube = self.shapes == ()
        self.ndims = self.cube.ndim

        self.name = escape(self.cube.name().title().replace("_", " "))
        self.names = [escape(dim_name) for dim_name in self._dim_names()]
        self.units = escape(str(self.cube.units))

    def _get_dim_names(self):
        """
        Get dimension-describing coordinate names, or '--' if no coordinate]
        describes the dimension.

        Note: borrows from `cube.summary`.

        """
        # Create a set to contain the axis names for each data dimension.
        dim_names = list(range(len(self.cube.shape)))

        # Add the dim_coord names that participate in the associated data
        # dimensions.
        for dim in range(len(self.cube.shape)):
            dim_coords = self.cube.coords(
                contains_dimension=dim, dim_coords=True
            )
            if dim_coords:
                dim_names[dim] = dim_coords[0].name()
            else:
                dim_names[dim] = "--"
        return dim_names

    def _dim_names(self):
        if self.scalar_cube:
            dim_names = ["(scalar cube)"]
        else:
            dim_names = self._get_dim_names()
        return dim_names

    def _get_lines(self):
        return self.cube_str.split("\n")

    def _get_bits(self, bits):
        """
        Parse the body content (`bits`) of the cube string in preparation for
        being converted into table rows.

        """
        left_indent = re.split(r"\w+", bits[1])[0]

        # Get heading indices within the printout.
        start_inds = []
        for hdg in self.str_headings.keys():
            heading = "{}{}".format(left_indent, hdg)
            try:
                start_ind = bits.index(heading)
            except ValueError:
                continue
            else:
                start_inds.append(start_ind)
        # Mark the end of the file.
        start_inds.append(0)

        # Retrieve info for each heading from the printout.
        for i0, i1 in zip(start_inds[:-1], start_inds[1:]):
            str_heading_name = bits[i0].strip()
            if i1 != 0:
                content = bits[i0 + 1 : i1]
            else:
                content = bits[i0 + 1 :]
            self.str_headings[str_heading_name] = content

    def _make_header(self):
        """
        Make the table header. This is similar to the summary of the cube,
        but does not include dim shapes. These are included on the next table
        row down, and produced with `make_shapes_row`.

        """
        # Header row.
        tlc_template = (
            '<th class="iris iris-word-cell">{self.name} ({self.units})</th>'
        )
        top_left_cell = tlc_template.format(self=self)
        cells = ['<tr class="iris">', top_left_cell]
        for dim_name in self.names:
            cells.append(
                '<th class="iris iris-word-cell">{}</th>'.format(dim_name)
            )
        cells.append("</tr>")
        return "\n".join(cell for cell in cells)

    def _make_shapes_row(self):
        """Add a row to show data / dimensions shape."""
        title_cell = (
            '<td class="iris-word-cell iris-subheading-cell">Shape</td>'
        )
        cells = ['<tr class="iris">', title_cell]
        for shape in self.shapes:
            cells.append(
                '<td class="iris iris-inclusion-cell">{}</td>'.format(shape)
            )
        cells.append("</tr>")
        return "\n".join(cell for cell in cells)

    def _make_row(self, title, body=None, col_span=0):
        """
        Produce one row for the table body; i.e.
            <tr><td>Coord name</td><td>x</td><td>-</td>...</tr>

        `body` contains the content for each cell not in the left-most (title)
               column.
               If None, indicates this row is a title row (see below).
        `title` contains the row heading. If `body` is None, indicates
                that the row contains a sub-heading;
                e.g. 'Dimension coordinates:'.
        `col_span` indicates how many columns the string should span.

        """
        row = ['<tr class="iris">']
        template = "    <td{html_cls}>{content}</td>"
        if body is None:
            # This is a title row.
            # Strip off the trailing ':' from the title string.
            title = title.strip()[:-1]
            row.append(
                template.format(
                    html_cls=' class="iris-title iris-word-cell"',
                    content=title,
                )
            )
            # Add blank cells for the rest of the rows.
            for _ in range(self.ndims):
                row.append(
                    template.format(html_cls=' class="iris-title"', content="")
                )
        else:
            # This is not a title row.
            # Deal with name of coord/attr etc. first.
            sub_title = "\t{}".format(title)
            row.append(
                template.format(
                    html_cls=' class="iris-word-cell iris-subheading-cell"',
                    content=sub_title,
                )
            )
            # One further item or more than that?
            if col_span != 0:
                html_cls = ' class="{}" colspan="{}"'.format(
                    "iris-word-cell", col_span
                )
                row.append(template.format(html_cls=html_cls, content=body))
            else:
                # "Inclusion" - `x` or `-`.
                for itm in body:
                    row.append(
                        template.format(
                            html_cls=' class="iris-inclusion-cell"',
                            content=itm,
                        )
                    )
        row.append("</tr>")
        return row

    def _expand_last_cell(self, element, body):
        """Expand an element containing a cell by adding a new line."""
        split_point = element.index("</td>")
        element = element[:split_point] + "<br>" + body + element[split_point:]
        return element

    def _make_content(self):
        elements = []
        for k, v in self.str_headings.items():
            if v is not None:
                # Add the sub-heading title.
                elements.extend(self._make_row(k))
                for line in v:
                    # Add every other row in the sub-heading.
                    if k in self.dim_desc_coords:
                        body = re.findall(r"[\w-]+", line)
                        title = body.pop(0)
                        colspan = 0
                    elif k in self.two_cell_headers:
                        try:
                            split_point = line.index(":")
                        except ValueError:
                            # When a line exists in v without a ':', we expect
                            # that this is due to the value of some attribute
                            # containing multiple lines. We collect all these
                            # lines in the same cell.
                            body = line.strip()
                            # We choose the element containing the last cell
                            # in the last row.
                            element = elements[-2]
                            element = self._expand_last_cell(element, body)
                            elements[-2] = element
                            continue
                        else:
                            title = line[:split_point].strip()
                            body = line[split_point + 2 :].strip()
                        colspan = self.ndims
                    else:
                        title = line.strip()
                        body = ""
                        colspan = self.ndims
                    elements.extend(
                        self._make_row(title, body=body, col_span=colspan)
                    )
        return "\n".join(element for element in elements)

    def repr_html(self):
        """The `repr` interface for Jupyter."""
        # Deal with the header first.
        header = self._make_header()

        # Check if we have a scalar cube.
        if self.scalar_cube:
            shape = ""
            # We still need a single content column!
            self.ndims = 1
        else:
            shape = self._make_shapes_row()

        # Now deal with the rest of the content.
        lines = self._get_lines()
        # If we only have a single line `cube_str` we have no coords / attrs!
        # We need to handle this case specially.
        if len(lines) == 1:
            content = ""
        else:
            self._get_bits(lines)
            content = self._make_content()

        return self._template.format(
            header=header, id=self.cube_id, shape=shape, content=content
        )


class CubeListRepresentation:
    _template = """
<style>
    .accordion-{uid} {{
        color: var(--jp-ui-font-color2);
        background: var(--jp-layout-color2);
        cursor: pointer;
        padding: 10px;
        border: 1px solid var(--jp-border-color0);
        width: 100%;
        text-align: left;
        font-size: 14px;
        font-family: var(--jp-code-font-family);
        font-weight: normal;
        outline: none;
        transition: 0.4s;
    }}
    .active {{
        background: var(--jp-layout-color1);
        font-weight: 900;
    }}
    .accordion-{uid}.active {{
        border: 1px solid var(--jp-brand-color1) !important;
    }}
    .accordion-{uid}:hover {{
        box-shadow: var(--jp-input-box-shadow);
        border: 2px solid var(--jp-brand-color1);
    }}
    .panel-{uid} {{
        padding: 0 18px;
        margin-bottom: 5px;
        background-color: var(--jp-layout-color1);
        display: none;
        overflow: hidden;
        border: 1px solid var(--jp-brand-color2);
    }}
</style>
<script type="text/javascript">
    var accordion = document.getElementsByClassName("accordion-{uid}");
    var i;

    for (i = 0; i < accordion.length; i++) {{
        accordion[i].addEventListener("click", function() {{
            this.classList.toggle("active");

            var panel = this.nextElementSibling;
            if (panel.style.display === "block") {{
                panel.style.display = "none";
            }} else {{
                panel.style.display = "block";
            }}
        }});
    }}
</script>
{contents}
    """

    _accordian_panel = """
<button class="accordion-{uid}">{title}</button>
<div class="panel-{uid}">
    <p>{content}</p>
</div>
    """

    def __init__(self, cubelist):
        self.cubelist = cubelist
        self.cubelist_id = id(self.cubelist)

    def make_content(self):
        html = []
        for i, cube in enumerate(self.cubelist):
            title = "{i}: {summary}".format(
                i=i, summary=cube.summary(shorten=True)
            )
            title = escape(title)
            content = cube._repr_html_()
            html.append(
                self._accordian_panel.format(
                    uid=self.cubelist_id, title=title, content=content
                )
            )
        return html

    def repr_html(self):
        contents = self.make_content()
        contents_str = "\n".join(contents)
        return self._template.format(
            uid=self.cubelist_id, contents=contents_str
        )



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the `iris._constraints.NameConstraint` class."""

from __future__ import absolute_import, division, print_function
from six.moves import filter, input, map, range, zip  # noqa

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from unittest.mock import Mock, sentinel

from iris._constraints import NameConstraint


class Test___init__(tests.IrisTest):
    def setUp(self):
        self.default = "none"

    def test_default(self):
        constraint = NameConstraint()
        self.assertEqual(constraint.standard_name, self.default)
        self.assertEqual(constraint.long_name, self.default)
        self.assertEqual(constraint.var_name, self.default)
        self.assertEqual(constraint.STASH, self.default)

    def test_standard_name(self):
        standard_name = sentinel.standard_name
        constraint = NameConstraint(standard_name=standard_name)
        self.assertEqual(constraint.standard_name, standard_name)
        constraint = NameConstraint(standard_name=standard_name)
        self.assertEqual(constraint.standard_name, standard_name)

    def test_long_name(self):
        long_name = sentinel.long_name
        constraint = NameConstraint(long_name=long_name)
        self.assertEqual(constraint.standard_name, self.default)
        self.assertEqual(constraint.long_name, long_name)
        constraint = NameConstraint(standard_name=None, long_name=long_name)
        self.assertIsNone(constraint.standard_name)
        self.assertEqual(constraint.long_name, long_name)

    def test_var_name(self):
        var_name = sentinel.var_name
        constraint = NameConstraint(var_name=var_name)
        self.assertEqual(constraint.standard_name, self.default)
        self.assertEqual(constraint.long_name, self.default)
        self.assertEqual(constraint.var_name, var_name)
        constraint = NameConstraint(
            standard_name=None, long_name=None, var_name=var_name
        )
        self.assertIsNone(constraint.standard_name)
        self.assertIsNone(constraint.long_name)
        self.assertEqual(constraint.var_name, var_name)

    def test_STASH(self):
        STASH = sentinel.STASH
        constraint = NameConstraint(STASH=STASH)
        self.assertEqual(constraint.standard_name, self.default)
        self.assertEqual(constraint.long_name, self.default)
        self.assertEqual(constraint.var_name, self.default)
        self.assertEqual(constraint.STASH, STASH)
        constraint = NameConstraint(
            standard_name=None, long_name=None, var_name=None, STASH=STASH
        )
        self.assertIsNone(constraint.standard_name)
        self.assertIsNone(constraint.long_name)
        self.assertIsNone(constraint.var_name)
        self.assertEqual(constraint.STASH, STASH)


class Test__cube_func(tests.IrisTest):
    def setUp(self):
        self.standard_name = sentinel.standard_name
        self.long_name = sentinel.long_name
        self.var_name = sentinel.var_name
        self.STASH = sentinel.STASH
        self.cube = Mock(
            standard_name=self.standard_name,
            long_name=self.long_name,
            var_name=self.var_name,
            attributes=dict(STASH=self.STASH),
        )

    def test_standard_name(self):
        # Match.
        constraint = NameConstraint(standard_name=self.standard_name)
        self.assertTrue(constraint._cube_func(self.cube))
        # Match.
        constraint = NameConstraint(standard_name=self.standard_name)
        self.assertTrue(constraint._cube_func(self.cube))
        # No match.
        constraint = NameConstraint(standard_name="wibble")
        self.assertFalse(constraint._cube_func(self.cube))
        # No match.
        constraint = NameConstraint(standard_name="wibble")
        self.assertFalse(constraint._cube_func(self.cube))

    def test_long_name(self):
        # Match.
        constraint = NameConstraint(long_name=self.long_name)
        self.assertTrue(constraint._cube_func(self.cube))
        # Match.
        constraint = NameConstraint(
            standard_name=self.standard_name, long_name=self.long_name
        )
        self.assertTrue(constraint._cube_func(self.cube))
        # No match.
        constraint = NameConstraint(long_name=None)
        self.assertFalse(constraint._cube_func(self.cube))
        # No match.
        constraint = NameConstraint(
            standard_name=None, long_name=self.long_name
        )
        self.assertFalse(constraint._cube_func(self.cube))

    def test_var_name(self):
        # Match.
        constraint = NameConstraint(var_name=self.var_name)
        self.assertTrue(constraint._cube_func(self.cube))
        # Match.
        constraint = NameConstraint(
            standard_name=self.standard_name,
            long_name=self.long_name,
            var_name=self.var_name,
        )
        self.assertTrue(constraint._cube_func(self.cube))
        # No match.
        constraint = NameConstraint(var_name=None)
        self.assertFalse(constraint._cube_func(self.cube))
        # No match.
        constraint = NameConstraint(
            standard_name=None, long_name=None, var_name=self.var_name
        )
        self.assertFalse(constraint._cube_func(self.cube))

    def test_STASH(self):
        # Match.
        constraint = NameConstraint(STASH=self.STASH)
        self.assertTrue(constraint._cube_func(self.cube))
        # Match.
        constraint = NameConstraint(
            standard_name=self.standard_name,
            long_name=self.long_name,
            var_name=self.var_name,
            STASH=self.STASH,
        )
        self.assertTrue(constraint._cube_func(self.cube))
        # No match.
        constraint = NameConstraint(STASH=None)
        self.assertFalse(constraint._cube_func(self.cube))
        # No match.
        constraint = NameConstraint(
            standard_name=None, long_name=None, var_name=None, STASH=self.STASH
        )
        self.assertFalse(constraint._cube_func(self.cube))


class Test___repr__(tests.IrisTest):
    def setUp(self):
        self.standard_name = sentinel.standard_name
        self.long_name = sentinel.long_name
        self.var_name = sentinel.var_name
        self.STASH = sentinel.STASH
        self.msg = "NameConstraint({})"
        self.f_standard_name = "standard_name={!r}".format(self.standard_name)
        self.f_long_name = "long_name={!r}".format(self.long_name)
        self.f_var_name = "var_name={!r}".format(self.var_name)
        self.f_STASH = "STASH={!r}".format(self.STASH)

    def test(self):
        constraint = NameConstraint()
        expected = self.msg.format("")
        self.assertEqual(repr(constraint), expected)

    def test_standard_name(self):
        constraint = NameConstraint(standard_name=self.standard_name)
        expected = self.msg.format(self.f_standard_name)
        self.assertEqual(repr(constraint), expected)

    def test_long_name(self):
        constraint = NameConstraint(long_name=self.long_name)
        expected = self.msg.format(self.f_long_name)
        self.assertEqual(repr(constraint), expected)
        constraint = NameConstraint(
            standard_name=self.standard_name, long_name=self.long_name
        )
        args = "{}, {}".format(self.f_standard_name, self.f_long_name)
        expected = self.msg.format(args)
        self.assertEqual(repr(constraint), expected)

    def test_var_name(self):
        constraint = NameConstraint(var_name=self.var_name)
        expected = self.msg.format(self.f_var_name)
        self.assertEqual(repr(constraint), expected)
        constraint = NameConstraint(
            standard_name=self.standard_name,
            long_name=self.long_name,
            var_name=self.var_name,
        )
        args = "{}, {}, {}".format(
            self.f_standard_name, self.f_long_name, self.f_var_name
        )
        expected = self.msg.format(args)
        self.assertEqual(repr(constraint), expected)

    def test_STASH(self):
        constraint = NameConstraint(STASH=self.STASH)
        expected = self.msg.format(self.f_STASH)
        self.assertEqual(repr(constraint), expected)
        constraint = NameConstraint(
            standard_name=self.standard_name,
            long_name=self.long_name,
            var_name=self.var_name,
            STASH=self.STASH,
        )
        args = "{}, {}, {}, {}".format(
            self.f_standard_name,
            self.f_long_name,
            self.f_var_name,
            self.f_STASH,
        )
        expected = self.msg.format(args)
        self.assertEqual(repr(constraint), expected)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test the cf module.

"""

# import iris tests first so that some things can be initialised before importing anything else
import iris.tests as tests

from unittest import mock

import iris
import iris.fileformats.cf as cf


class TestCaching(tests.IrisTest):
    def test_cached(self):
        # Make sure attribute access to the underlying netCDF4.Variable
        # is cached.
        name = "foo"
        nc_var = mock.MagicMock()
        cf_var = cf.CFAncillaryDataVariable(name, nc_var)
        self.assertEqual(nc_var.ncattrs.call_count, 1)

        # Accessing a netCDF attribute should result in no further calls
        # to nc_var.ncattrs() and the creation of an attribute on the
        # cf_var.
        # NB. Can't use hasattr() because that triggers the attribute
        # to be created!
        self.assertTrue("coordinates" not in cf_var.__dict__)
        _ = cf_var.coordinates
        self.assertEqual(nc_var.ncattrs.call_count, 1)
        self.assertTrue("coordinates" in cf_var.__dict__)

        # Trying again results in no change.
        _ = cf_var.coordinates
        self.assertEqual(nc_var.ncattrs.call_count, 1)
        self.assertTrue("coordinates" in cf_var.__dict__)

        # Trying another attribute results in just a new attribute.
        self.assertTrue("standard_name" not in cf_var.__dict__)
        _ = cf_var.standard_name
        self.assertEqual(nc_var.ncattrs.call_count, 1)
        self.assertTrue("standard_name" in cf_var.__dict__)


@tests.skip_data
class TestCFReader(tests.IrisTest):
    def setUp(self):
        filename = tests.get_data_path(
            ("NetCDF", "rotated", "xyt", "small_rotPole_precipitation.nc")
        )
        self.cfr = cf.CFReader(filename)

    def test_ancillary_variables_pass_0(self):
        self.assertEqual(self.cfr.cf_group.ancillary_variables, {})

    def test_auxiliary_coordinates_pass_0(self):
        self.assertEqual(
            sorted(self.cfr.cf_group.auxiliary_coordinates.keys()),
            ["lat", "lon"],
        )

        lat = self.cfr.cf_group["lat"]
        self.assertEqual(lat.shape, (190, 174))
        self.assertEqual(lat.dimensions, ("rlat", "rlon"))
        self.assertEqual(lat.ndim, 2)
        self.assertEqual(
            lat.cf_attrs(),
            (
                ("long_name", "latitude"),
                ("standard_name", "latitude"),
                ("units", "degrees_north"),
            ),
        )

        lon = self.cfr.cf_group["lon"]
        self.assertEqual(lon.shape, (190, 174))
        self.assertEqual(lon.dimensions, ("rlat", "rlon"))
        self.assertEqual(lon.ndim, 2)
        self.assertEqual(
            lon.cf_attrs(),
            (
                ("long_name", "longitude"),
                ("standard_name", "longitude"),
                ("units", "degrees_east"),
            ),
        )

    def test_bounds_pass_0(self):
        self.assertEqual(
            sorted(self.cfr.cf_group.bounds.keys()), ["time_bnds"]
        )

        time_bnds = self.cfr.cf_group["time_bnds"]
        self.assertEqual(time_bnds.shape, (4, 2))
        self.assertEqual(time_bnds.dimensions, ("time", "time_bnds"))
        self.assertEqual(time_bnds.ndim, 2)
        self.assertEqual(time_bnds.cf_attrs(), ())

    def test_coordinates_pass_0(self):
        self.assertEqual(
            sorted(self.cfr.cf_group.coordinates.keys()),
            ["rlat", "rlon", "time"],
        )

        rlat = self.cfr.cf_group["rlat"]
        self.assertEqual(rlat.shape, (190,))
        self.assertEqual(rlat.dimensions, ("rlat",))
        self.assertEqual(rlat.ndim, 1)
        attr = []
        attr.append(("axis", "Y"))
        attr.append(("long_name", "rotated latitude"))
        attr.append(("standard_name", "grid_latitude"))
        attr.append(("units", "degrees"))
        self.assertEqual(rlat.cf_attrs(), tuple(attr))

        rlon = self.cfr.cf_group["rlon"]
        self.assertEqual(rlon.shape, (174,))
        self.assertEqual(rlon.dimensions, ("rlon",))
        self.assertEqual(rlon.ndim, 1)
        attr = []
        attr.append(("axis", "X"))
        attr.append(("long_name", "rotated longitude"))
        attr.append(("standard_name", "grid_longitude"))
        attr.append(("units", "degrees"))
        self.assertEqual(rlon.cf_attrs(), tuple(attr))

        time = self.cfr.cf_group["time"]
        self.assertEqual(time.shape, (4,))
        self.assertEqual(time.dimensions, ("time",))
        self.assertEqual(time.ndim, 1)
        attr = []
        attr.append(("axis", "T"))
        attr.append(("bounds", "time_bnds"))
        attr.append(("calendar", "gregorian"))
        attr.append(("long_name", "Julian Day"))
        attr.append(("units", "days since 1950-01-01 00:00:00.0"))
        self.assertEqual(time.cf_attrs(), tuple(attr))

    def test_data_pass_0(self):
        self.assertEqual(
            sorted(self.cfr.cf_group.data_variables.keys()), ["pr"]
        )

        data = self.cfr.cf_group["pr"]
        self.assertEqual(data.shape, (4, 190, 174))
        self.assertEqual(data.dimensions, ("time", "rlat", "rlon"))
        self.assertEqual(data.ndim, 3)
        attr = []
        attr.append(("_FillValue", 1e30))
        attr.append(("cell_methods", "time: mean"))
        attr.append(("coordinates", "lon lat"))
        attr.append(("grid_mapping", "rotated_pole"))
        attr.append(("long_name", "Precipitation"))
        attr.append(("missing_value", 1e30))
        attr.append(("standard_name", "precipitation_flux"))
        attr.append(("units", "kg m-2 s-1"))
        attr = tuple(attr)
        self.assertEqual(data.cf_attrs()[0][0], attr[0][0])
        self.assertAlmostEqual(data.cf_attrs()[0][1], attr[0][1], delta=1.6e22)
        self.assertEqual(data.cf_attrs()[1:5], attr[1:5])
        self.assertAlmostEqual(data.cf_attrs()[5][1], attr[5][1], delta=1.6e22)
        self.assertEqual(data.cf_attrs()[6:], attr[6:])

    def test_formula_terms_pass_0(self):
        self.assertEqual(self.cfr.cf_group.formula_terms, {})

    def test_grid_mapping_pass_0(self):
        self.assertEqual(
            sorted(self.cfr.cf_group.grid_mappings.keys()), ["rotated_pole"]
        )

        rotated_pole = self.cfr.cf_group["rotated_pole"]
        self.assertEqual(rotated_pole.shape, ())
        self.assertEqual(rotated_pole.dimensions, ())
        self.assertEqual(rotated_pole.ndim, 0)
        attr = []
        attr.append(("grid_mapping_name", "rotated_latitude_longitude"))
        attr.append(("grid_north_pole_latitude", 18.0))
        attr.append(("grid_north_pole_longitude", -140.75))
        self.assertEqual(rotated_pole.cf_attrs(), tuple(attr))

    def test_cell_measures_pass_0(self):
        self.assertEqual(self.cfr.cf_group.cell_measures, {})

    def test_global_attributes_pass_0(self):
        self.assertEqual(
            sorted(self.cfr.cf_group.global_attributes.keys()),
            [
                "Conventions",
                "NCO",
                "experiment",
                "history",
                "institution",
                "source",
            ],
        )

        self.assertEqual(
            self.cfr.cf_group.global_attributes["Conventions"], "CF-1.0"
        )
        self.assertEqual(
            self.cfr.cf_group.global_attributes["experiment"], "ER3"
        )
        self.assertEqual(
            self.cfr.cf_group.global_attributes["institution"], "DMI"
        )
        self.assertEqual(
            self.cfr.cf_group.global_attributes["source"], "HIRHAM"
        )

    def test_variable_cf_group_pass_0(self):
        self.assertEqual(
            sorted(self.cfr.cf_group["time"].cf_group.keys()), ["time_bnds"]
        )
        self.assertEqual(
            sorted(self.cfr.cf_group["pr"].cf_group.keys()),
            ["lat", "lon", "rlat", "rlon", "rotated_pole", "time"],
        )

    def test_variable_attribute_touch_pass_0(self):
        lat = self.cfr.cf_group["lat"]

        self.assertEqual(
            lat.cf_attrs(),
            (
                ("long_name", "latitude"),
                ("standard_name", "latitude"),
                ("units", "degrees_north"),
            ),
        )
        self.assertEqual(lat.cf_attrs_used(), ())
        self.assertEqual(
            lat.cf_attrs_unused(),
            (
                ("long_name", "latitude"),
                ("standard_name", "latitude"),
                ("units", "degrees_north"),
            ),
        )

        # touch some variable attributes.
        lat.long_name
        lat.units
        self.assertEqual(
            lat.cf_attrs_used(),
            (("long_name", "latitude"), ("units", "degrees_north")),
        )
        self.assertEqual(
            lat.cf_attrs_unused(), (("standard_name", "latitude"),)
        )

        # clear the attribute touch history.
        lat.cf_attrs_reset()
        self.assertEqual(lat.cf_attrs_used(), ())
        self.assertEqual(
            lat.cf_attrs_unused(),
            (
                ("long_name", "latitude"),
                ("standard_name", "latitude"),
                ("units", "degrees_north"),
            ),
        )


@tests.skip_data
class TestLoad(tests.IrisTest):
    def test_attributes_empty(self):
        filename = tests.get_data_path(
            ("NetCDF", "global", "xyt", "SMALL_hires_wind_u_for_ipcc4.nc")
        )
        cube = iris.load_cube(filename)
        self.assertEqual(cube.coord("time").attributes, {})

    def test_attributes_contain_positive(self):
        filename = tests.get_data_path(
            ("NetCDF", "global", "xyt", "SMALL_hires_wind_u_for_ipcc4.nc")
        )
        cube = iris.load_cube(filename)
        self.assertEqual(cube.coord("height").attributes["positive"], "up")

    def test_attributes_populated(self):
        filename = tests.get_data_path(
            ("NetCDF", "label_and_climate", "small_FC_167_mon_19601101.nc")
        )
        cube = iris.load_cube(filename, "air_temperature")
        self.assertEqual(
            sorted(cube.coord("longitude").attributes.items()),
            [
                ("data_type", "float"),
                ("modulo", 360),
                ("topology", "circular"),
                ("valid_max", 359.0),
                ("valid_min", 0.0),
            ],
        )

    def test_cell_methods(self):
        filename = tests.get_data_path(
            ("NetCDF", "global", "xyt", "SMALL_hires_wind_u_for_ipcc4.nc")
        )
        cube = iris.load_cube(filename)
        self.assertEqual(
            cube.cell_methods,
            (
                iris.coords.CellMethod(
                    method="mean",
                    coords=("time",),
                    intervals=("6 minutes",),
                    comments=(),
                ),
            ),
        )


@tests.skip_data
class TestClimatology(tests.IrisTest):
    def setUp(self):
        filename = tests.get_data_path(
            (
                "NetCDF",
                "label_and_climate",
                "A1B-99999a-river-sep-2070-2099.nc",
            )
        )
        self.cfr = cf.CFReader(filename)

    def test_bounds(self):
        time = self.cfr.cf_group["temp_dmax_tmean_abs"].cf_group.coordinates[
            "time"
        ]
        climatology = time.cf_group.climatology
        self.assertEqual(len(climatology), 1)
        self.assertEqual(list(climatology.keys()), ["climatology_bounds"])

        climatology_var = climatology["climatology_bounds"]
        self.assertEqual(climatology_var.ndim, 2)
        self.assertEqual(climatology_var.shape, (1, 2))


@tests.skip_data
class TestLabels(tests.IrisTest):
    def setUp(self):
        filename = tests.get_data_path(
            (
                "NetCDF",
                "label_and_climate",
                "A1B-99999a-river-sep-2070-2099.nc",
            )
        )
        self.cfr_start = cf.CFReader(filename)

        filename = tests.get_data_path(
            ("NetCDF", "label_and_climate", "small_FC_167_mon_19601101.nc")
        )
        self.cfr_end = cf.CFReader(filename)

    def test_label_dim_start(self):
        cf_data_var = self.cfr_start.cf_group["temp_dmax_tmean_abs"]

        region_group = self.cfr_start.cf_group.labels["region_name"]
        self.assertEqual(
            sorted(self.cfr_start.cf_group.labels.keys()), ["region_name"]
        )
        self.assertEqual(
            sorted(cf_data_var.cf_group.labels.keys()), ["region_name"]
        )

        self.assertEqual(
            region_group.cf_label_dimensions(cf_data_var), ("georegion",)
        )
        self.assertEqual(region_group.cf_label_data(cf_data_var)[0], "Anglian")

        cf_data_var = self.cfr_start.cf_group["cdf_temp_dmax_tmean_abs"]

        self.assertEqual(
            sorted(self.cfr_start.cf_group.labels.keys()), ["region_name"]
        )
        self.assertEqual(
            sorted(cf_data_var.cf_group.labels.keys()), ["region_name"]
        )

        self.assertEqual(
            region_group.cf_label_dimensions(cf_data_var), ("georegion",)
        )
        self.assertEqual(region_group.cf_label_data(cf_data_var)[0], "Anglian")

    def test_label_dim_end(self):
        cf_data_var = self.cfr_end.cf_group["tas"]

        self.assertEqual(
            sorted(self.cfr_end.cf_group.labels.keys()),
            ["experiment_id", "institution", "source"],
        )
        self.assertEqual(
            sorted(cf_data_var.cf_group.labels.keys()),
            ["experiment_id", "institution", "source"],
        )

        self.assertEqual(
            self.cfr_end.cf_group.labels["experiment_id"].cf_label_dimensions(
                cf_data_var
            ),
            ("ensemble",),
        )
        self.assertEqual(
            self.cfr_end.cf_group.labels["experiment_id"].cf_label_data(
                cf_data_var
            )[0],
            "2005",
        )

        self.assertEqual(
            self.cfr_end.cf_group.labels["institution"].cf_label_dimensions(
                cf_data_var
            ),
            ("ensemble",),
        )
        self.assertEqual(
            self.cfr_end.cf_group.labels["institution"].cf_label_data(
                cf_data_var
            )[0],
            "ECMWF",
        )

        self.assertEqual(
            self.cfr_end.cf_group.labels["source"].cf_label_dimensions(
                cf_data_var
            ),
            ("ensemble",),
        )
        self.assertEqual(
            self.cfr_end.cf_group.labels["source"].cf_label_data(cf_data_var)[
                0
            ],
            "IFS33R1/HOPE-E, Sys 1, Met 1, ENSEMBLES",
        )


if __name__ == "__main__":
    tests.main()
